!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
 
 module setio
  use struct
  use parallel
  use griddata
  use settings
  use particles

  implicit none
  !--------------------------
  integer,parameter :: sz_par=5
  integer,parameter :: SUMV=0,MINv=1,MAXV=2
  real(kind=4):: rpar(sz_par)
  real  :: par(sz_par)
  integer:: k,ipar(sz_par)
  integer,parameter :: ne=128
  integer :: nbd_tot
                                !clock parameters for outputs
  real :: dt_out,tout,dt_ene,te
  integer :: iout,iene,nout,ite,nte

  real,allocatable :: enfun(:,:),fdes(:,:)
  real :: entime(200)
  type(particle),allocatable :: pd(:)
  !--------------------------------
  contains
!===========================
! Allocate particles position-velocity and acceleration arrays
!===========================================
  subroutine nb_alloc(nb,psize)
   integer,intent(in) :: nb
   integer,intent(inout) :: psize 

   allocate(xvp(1:nb,0:6))
   allocate(p_acc(1:nb,0:3))
   xvp=0.0
   p_acc=0.0
   psize=psize+size(xvp)+size(p_acc)

  end subroutine nb_alloc
!===========================
! Reads particles input parameters and initial position-velocities
!===========================================
 subroutine read_data_par(id_new)
                                   ! exit with rpar and ipar in pe0
   integer,intent(in)  :: id_new
   character(6) :: fname='      '

     if (id_new<100) write (fname,'(a4,i2)') 'mout' ,id_new
     if (id_new< 10) write (fname,'(a5,i1)') 'mout0',id_new
     open(10,file=fname//'.bin',form='unformatted')

     read(10)ipar(1:5)
     read(10)rpar(1:5)

     close(10)
!
  end subroutine read_data_par
!===========================
! Reads particles initial position-velocities
! Exit with particles data distributed on each PE 
!===========================================
 subroutine read_data(id_new)
   character(6) :: fname='      '
   integer:: id_new,k,k1,k2,ipe,nb_loc
   real(kind=4)  :: var(6)
    real :: x0(3),x1(3)
   logical :: sr

   allocate(pd(nbd_max)) !auxiliary structure array to exchage data

    if(pe0)then
     if (id_new<100) write (fname,'(a4,i2)') 'mout' ,id_new
     if (id_new< 10) write (fname,'(a5,i1)') 'mout0',id_new
     write(99,*)'pe0 reads data: '//fname//'.bin'
     open(10,file=fname//'.bin',form='unformatted')

     read(10)ipar(1:5)
     read(10)rpar(1:5)
!
     do k=1,nbd(0)
      read(10)var(1:6)
      xvp(k,1:6)=var(1:6)
     end do
    endif
!------------
    if(prl)then
     if(pe0)then
      sr=.true.
      k1=nbd(0)+1
      do ipe=1,pe_max
       nb_loc=nbd(ipe)
       do k=1,nb_loc
        read(10)pd(k)%posvel(1:6)
       end do
       call exchange_pdata(sr,pd(1:nb_loc),nb_loc,ipe,ipe+1)
      end do
     else
      nb_loc=nbd(mype)
      sr=.false.
      call exchange_pdata(sr,pd(1:nb_loc),nb_loc,pe_min,mype+1)
      do k=1,nb_loc
       xvp(k,1:6)=pd(k)%posvel(1:6)
      end do
     endif
   endif             ! end of prl
   if(allocated(pd))deallocate(pd)

   if(pe0)close(10)
   nb_loc=nbd(mype)
!----------------- reset COM
   do k=1,3
    x0(k)=pmass*sum(xvp(1:nb_loc,k))
   end do 
   call allreduce_real_data(3,x0,x1,SUMV)

  end subroutine read_data
!==============================================
!Initializes by reading and storing particles data
!Implemented models for initial conditions:
!1) Equilibrium gamma=1
!2) Equilibrium Plummer 
!3) Cold  Plummer 
!4) Cold  gamma=1
!==============================================
  subroutine part_init(id_new,psize)
    integer,intent(in):: id_new
    integer,intent(out) :: psize
    integer :: sizeall,imods
    character(20),dimension(4) :: model=(/'Equilib.gamma-model','Equilibrium Plummer',& 
    '   Cold Plummer    ',' Cold gamma-model  '/) 


    if(pe0)then
     write(99,*)'  READ PARTICLE DATA'
     call read_data_par(id_new)       ! first read N_data (rpar,ipar)
     par=rpar
    endif
    if(prl)then
     call bcast_real_data(sz_par,par)
     call bcast_int_data(sz_par,ipar)
    endif

    nbd_tot=ipar(1)
    imods=ipar(2)
    totm=par(1)
    pmass=totm/float(nbd_tot)
!-----------------------
    tnow=par(2)
    tdyn=par(3)
    vir_0=par(4)
!------------------------
    if(nbd_tot==0)then
     ier=8
     return
    endif
    nbd(0)=nbd_tot/npe
    nbd_max=nbd(0)
    if(prl)nbd(1:npe-1)=nbd(0)

   call nb_alloc(nbd(mype),psize)           !
   !--------------------------------------------------------
   call read_data(id_new)             ! reads particles array

   iout=id_new
!
   if(pe0)then
    write(99,*)'Initial condition of the running model: '
    write(99,*)model(imods)
    write(99,*)'****************************************'
    
    write(99,*)' PARTICLE DATA '
    write(99,'(a11,2f5.1)') 'tot mass = ',totm
    write(99,'(a15,2e11.4)') 'particle mass =',pmass
    write(99,'(a15,e11.4)')'initial time = ',tnow
    write(99,'(a17,e11.4)')'dynamical time = ',tdyn
    write(99,'(a20,e11.4)')'initial vir ratio = ',vir_0
    write(99,*)'****************************************'
   allocate(enfun(1000,25),fdes(-1:ne,100))
   enfun=0.0
   fdes=0.0
endif

  end subroutine part_init
!--------------
 subroutine pot_data
                                ! exit with pdata on each pe
   integer:: k,ipe
   real(kind=4),allocatable  :: var(:)
   logical :: sr
   integer :: k1,k2,nb_loc

   character(6) :: fname='     '

    allocate(var(nbd_max))

    if(pe0)then
     if (iout<100) write (fname,'(a4,i2)') 'pout' ,iout
     if (iout< 10) write (fname,'(a5,i1)') 'pout0',iout
     write(99,*)'pe0 writes particle data on : '//fname//'.bin'
     open(10,file=fname//'.bin',form='unformatted')

     
     rpar(2)=tnow
     rpar(3)=tdyn
     ipar(3)=mond_ind
     
     write(10)ipar(1:5)
     write(10)rpar(1:5)
     nb_loc=nbd(0)
     var(1:nb_loc)=p_acc(1:nb_loc,0)
     do k=1,nb_loc
      write(10)var(k)
     end do
    endif
!------------
    if(prl)then
     if(.not.pe0)then
      sr=.true.
      ipe=mype
      nb_loc=nbd(mype)
      call exchange_rdata(sr,p_acc(1:nb_loc,0),nb_loc,pe_min,ipe+1)
     else
      sr=.false.
      do ipe=1,pe_max
       nb_loc=nbd(ipe)
       call exchange_rdata(sr,p_acc(1:nb_loc,0),nb_loc,ipe,ipe+1)
       var(1:nb_loc)=p_acc(1:nb_loc,0)
       do k=1,nb_loc
        write(10)var(k)
       end do
      end do
     endif
   endif             ! end of prl
   if(pe0)close(10)
    if(allocated(var))deallocate(var)

  end subroutine pot_data 
!----------------------
 subroutine part_output
                                ! exit with pdata on each pe
   integer:: k,ipe
   real(kind=4)  :: var(6)
   logical :: sr
   integer :: k1,k2,nb_loc

   character(6) :: fname='     '

   iout=iout+1
    if(pe0)then
     if (iout<100) write (fname,'(a4,i2)') 'mout' ,iout
     if (iout< 10) write (fname,'(a5,i1)') 'mout0',iout
     write(99,*)'pe0 writes particle data on : '//fname//'.bin'
     open(10,file=fname//'.bin',form='unformatted')

     
     rpar(2)=tnow
     rpar(3)=tdyn
     ipar(3)=mond_ind
     
     write(10)ipar(1:5)
     write(10)rpar(1:5)
!
     do k=1,nbd(0)
      var(1:6)=xvp(k,1:6)
      write(10)var(1:6)
     end do
    endif
!------------
    if(prl)then
     allocate(pd(nbd_max))
     if(.not.pe0)then
      sr=.true.
      ipe=mype
      nb_loc=nbd(mype)
      do k=1,nb_loc
       pd(k)%posvel(1:6)=xvp(k,1:6)
      end do
      call exchange_pdata(sr,pd(1:nb_loc),nb_loc,pe_min,ipe+1)
     else
      sr=.false.
      do ipe=1,pe_max
       nb_loc=nbd(ipe)
       call exchange_pdata(sr,pd(1:nb_loc),nb_loc,ipe,ipe+1)
       do k=1,nb_loc
        write(10)pd(k)%posvel(1:6)
       end do
      end do
     endif
     if(allocated(pd))deallocate(pd)
   endif             ! end of prl
   if(pe0)close(10)

  end subroutine part_output 
!+++++++++++++++++++++++++++++++++++++
! To reset the input particles velociry scale to the nominal virial ratio 
!==============================
 subroutine vir_reset
   integer:: p,k
   real:: epot(2),epot1(2),mass0,vfact,ent,x0(3) 
   logical :: ib
!-----------------------
  epot=0.0
  mass0=pmass
  do p=1,nbd(mype)
   x0(1:3)=xvp(p,1:3)
   epot(1)=epot(1)+dot_product(x0(1:3),p_acc(p,1:3))
   epot(2)=epot(2)&
         +dot_product(xvp(p,4:6),xvp(p,4:6))
  end do
  call allreduce_real_data(2,epot1(1:2),epot(1:2),SUMV)
  vfact=1.0
  if(epot(2)> 0.0)vfact=sqrt(vir_0*abs(epot(1))/epot(2))
  do p=1,nbd(mype)
    xvp(p,4:6)=vfact*xvp(p,4:6)
  end do
  epot(2)=vfact*vfact*epot(2)
!----------------- now reset tdyn
  ent=mass0*abs(0.5*epot(2)-epot(1))       !  
  tdyn=abs(ent/totm)
  tdyn=(2.0*tdyn)*sqrt(2.0*tdyn)
  tdyn=totm/tdyn
  if(pe0)then
   write(99,*)' virial ratio reset'
   write(99,'(a7,e11.4)')'  tdyn=',tdyn
  endif
 end subroutine vir_reset
!++++++++++++++++++++++++++
! Writes grid based fields: density, potential and acceleration
! and the related parameters in a binary file 
!=========================================
 subroutine grid_bdata(gout,nph1)
  integer,intent(in) :: gout,nph1
  character(6) :: gname='      '
  integer :: i,j,nph2,nfields
  real(kind=4),allocatable :: fdata(:,:,:)
  real(kind=4) :: stnow,sg_0,srad_sc,srh
   
  write (gname,'(a5,i1)')'mond0',gout
  nph2=nph
  if(nph1<=4)nph2=1
  allocate(fdata(nr+1,nth,nph2))
  stnow=tnow
  sg_0=g_0
  srad_sc=rad_sc
  srh=rh

  open(10,file=gname//'.bin',form='unformatted')
  nfields=5

  write(10)nfields,nr,nth,nph,nph2,lmax,itermax,rmap,mond_ind,spl_ord
  write(10)stnow,sg_0,srad_sc,srh 
  fdata(1:nr+1,1,1)=r(0:nr,1)
  write(10)fdata(1:nr+1,1,1)
  fdata(1,1:nth,1)=th(0:nth-1)
  write(10)fdata(1,1:nth,1)
  fdata(1,1,1:nph2)=phi(0:nph2-1)
  write(10)fdata(1,1,1:nph2)
  
  do j=0,nth-1
   do i=0,nr
    fdata(i+1,j+1,1:nph2)=den(i,j,0:nph2-1)/(r(i,1)*gravc)
   enddo
  end do
  write(10)fdata
  fdata(1:nr+1,1:nth,1:nph2)=aux(0:nr,0:nth-1,0:nph2-1)
  write(10)fdata
  fdata(1:nr+1,1:nth,1:nph2)=Gr(0:nr,0:nth-1,0:nph2-1)
  write(10)fdata
  do j=0,nth-1
   do i=0,nr
    fdata(i+1,j+1,1:nph2)=Gth(i,j,0:nph2-1)/r(i,1)
   enddo
  end do
  write(10)fdata
  do j=0,nth-1
   do i=0,nr
    fdata(i+1,j+1,1:nph2)=Gph(i,j,0:nph2-1)/(r(i,1)*sth(j))
   enddo
  end do
  write(10)fdata
  close(10)
  if(allocated(fdata))deallocate(fdata)

  if(pe0)write(99,*)'pe0 writes grid data on : '//gname//'.bin'

 end subroutine grid_bdata
!++++++++++++++++++++++++++++++++
! Routines for computing integrated particles properties
!=================================== 
! Angular and linear momentum (COM)components
!====================================
 subroutine momenta(var)
   real,intent(out) :: var(6)
   integer :: ip,nb_loc
   real :: mass0,x(3),v(3)
   
   var=0.0
   mass0=pmass
   nb_loc=nbd(mype)
    do ip=1,nb_loc
     x(1:3)=xvp(ip,1:3)
     var(1:3)=var(1:3)+mass0*x(1:3)
     v(1:3)=xvp(ip,4:6)-hdt*p_acc(ip,1:3)
     
     var(4)=var(4)+pmass*(v(2)*x(3)-v(3)*x(2))
     var(5)=var(5)+pmass*(v(1)*x(3)-v(3)*x(1))
     var(6)=var(6)+pmass*(v(2)*x(1)-v(1)*x(2))

   end do
  end subroutine momenta 
!=======================
! Kinetic energy components (velocity dispersion) in spherical coordunates
! Integrated inside a sphere of prescribed r0 input value
!=================================
 subroutine Ek_cmp(r0,var)
   real,intent(in) :: r0 
   real,intent(out) :: var(3)
   integer :: ip,nb_loc
   real :: mass0,x(3),v(3),rad,r2,rs,vr,vth,vph,v2
   
   var=0.0
   mass0=pmass
   nb_loc=nbd(mype)
    do ip=1,nb_loc
      
     x(1:3)=xvp(ip,1:3)
     rs=x(1)*x(1)+x(2)*x(2)
     r2=rs+x(3)*x(3)
     rs=sqrt(rs)
     rad=sqrt(r2)
     if(rad <=r0)then
      v(1:3)=xvp(ip,4:6)-hdt*p_acc(ip,1:3)
      v2=dot_product(v(1:3),v(1:3))
      vr=dot_product(x(1:3),v(1:3))/rad
      vph=(x(1)*v(2)-x(2)*v(1))/rs
      vph=vph*vph
      vr=vr*vr
      vth=v2-vr-vph
      var(1)=var(1)+mass0*vr
      var(2)=var(2)+mass0*vth
      var(3)=var(3)+mass0*vph
     endif
   end do
   var(1:3)=0.5*var(1:3)
 end subroutine Ek_cmp 
!==========================
! Total kinetic energy and potential energy of particles
! Inertia tensor components inside a sphere of prescribed r0 radius
!=========================================
 subroutine entens(r0,tv,tx,ep)

   real,intent(in) :: r0
   real,intent(out) :: tv(6),tx(0:6),ep
    integer  ::  ip
    real :: x0(3),x1(3),v0(3),rad,mass0

    tv=0.0
    tx=0.0
    mass0=pmass
    ep=0.0
     
     do ip=1,nbd(mype)
      v0(1:3)=xvp(ip,4:6)-hdt*p_acc(ip,1:3) !velocities at n+1/2 time level
      v0(1:3)=v0(1:3)*v0(1:3)
      x0(1:3)=xvp(ip,1:3)
      x1(1:3)=x0(1:3)-hdt*xvp(ip,4:6)       !positions at n+1/2 time level
      ep=ep+p_acc(ip,0)

      tv(1:3)=tv(1:3)+v0(1:3)
      tx(0)=tx(0)+dot_product(x1(1:3),p_acc(ip,1:3))

      rad=dot_product(x0(1:3),x0(1:3))
      rad=sqrt(rad)
      if(rad <=r0)then 
       tx(1:3)=tx(1:3)+x0(1:3)*x0(1:3)
       tx(4)=tx(4)+x0(1)*x0(2)
       tx(5)=tx(5)+x0(1)*x0(3)
       tx(6)=tx(6)+x0(2)*x0(3)
      endif
     end do
     ep=0.5*ep*mass0
     tx=tx*mass0
     tv(1:3)=tv(1:3)*mass0
  end subroutine entens
!===========================
! The particles distribution along the energy axis
!=========================== 
  subroutine NdE(it)
   integer,intent(in) :: it
   real :: en,emax,emax1,emin,emin1,de,v0(3),ekin,mass0,ax
   real :: fde(130),fde1(130)
   integer :: ip,idp,kh,k
   logical :: ib


   fde(:)=0.0
   emin=10000.
   emax=-emin
   do ip=1,nbd(mype)
    v0(1:3)=xvp(ip,4:6)-hdt*p_acc(ip,1:3)
    ekin=0.5*dot_product(v0(1:3),v0(1:3))
    en=p_acc(ip,0)-ekin
    emax=max(en,emax)
    emin=min(en,emin)
   end do
    call allreduce_real_sdata(emax1,emax,MAXV)
    call allreduce_real_sdata(emin1,emin,MINV)

   de=(emax-emin)/float(ne)
   mass0=pmass

    do ip=1,nbd(mype)
     v0(1:3)=xvp(ip,4:6)-hdt*p_acc(ip,1:3)
     ekin=0.5*dot_product(v0(1:3),v0(1:3))
     en=p_acc(ip,0)-ekin
     ax=(en-emin)/de
     k=ax
     ax=ax-float(k)
     k=k+1
     if(k> 0.and.k < ne)then
      fde(k)=fde(k)+mass0*(1.0-ax)
      fde(k+1)=fde(k+1)+mass0*ax
     endif
    enddo
    fde1=fde
    call allreduce_real_data(ne,fde(1:ne),fde1(1:ne),SUMV)
    if(pe0)then
     fdes(-1,it)=emin
     fdes(0,it)=emax
     fdes(1:ne,it)=fde1(1:ne)
    endif
  end subroutine NdE
!===================================
! Computes and stores for later output the time sequence of 
! total energy components,,of density max,D_max, of the velocity dispersion
! and Inertia tensor components, and of total linear and angular momenta
!===================================
  subroutine diags
    integer::  ip,kh
    real     :: amvec(9),amvec1(9),tx(0:6),tv(6)
    real     :: ekin,epot
    logical :: ib

    ite=ite+1
    ib=.false.             !no broacast needed

                                 !  W, ekin, Inertia tensor
    amvec=0.0
                     !pot_max=maxval(p_acc(1:nbd(mype),0))
                     !print *,'epot',0.5*pmass*sum(p_acc(1:nbd(mype),0))
     call entens(rh,tv,tx,epot)
     amvec(1)=-tx(0)
     amvec(2)=0.5*sum(tv(1:3))
     amvec(3:8)=0.5*tx(1:6)
     amvec(9)=epot

     call allreduce_real_data(9,amvec(1:9),amvec1(1:9),SUMV)

    amvec(1)=epf
    amvec(2)=dmax
    amvec(3)=rh
    if(pe0)then
     enfun(ite,1)=tnow
     enfun(ite,2:3)=amvec1(1:2)          !E_int,E_kin
     enfun(ite,4)=amvec1(9)              !Epot
     enfun(ite,5:7)=amvec(1:3)           !Enf D-max, R_h
     enfun(ite,17:22)=amvec1(3:8)        ! inertia tensor      
    end if
    amvec=0.0

    call Ek_cmp(rh,amvec(1:3))
    call momenta(amvec(4:9))

    call allreduce_real_data(9,amvec(1:9),amvec1(1:9),SUMV)
    if(pe0)then
     enfun(ite,8:16)=amvec1(1:9)
    end if

   te=te+dt_ene
!--------------------------------------
                                         !N(E)dE: NOT COMPUTED 
!    if(mod(ite,10)==0)then
!     nte=nte+1
!     entime(nte)=tnow
!     call NdE(nte)
!    endif

 end subroutine diags
!=======================
! Writes on a ASCII file the time sequence of the stored integrated quantities
!============================== 
  subroutine endata(itg,itnde,io)
   integer,intent(in) :: itg,itnde,io
   integer,parameter:: lun=40
   character(6) :: fname='      '
   integer:: i,k,envar,nbd1,itg_old

   character(5),dimension(22) :: varlb=(/' Time','Eint ','Ekin ',&
'Epot ',' Enf ',&
'D_max',' R_h ','Ek_r ','Ek_th','Ek_ph',' P_x ',' P_y ',' P_z ',&
' L_x ',' L_y ',' L_z ',&
' T_xx',' T_yy',' T_zz',' T_xy',' T_xz',' T_yz'/)

!
  if(pe0)then
   if (io<100) write (fname,'(a4,i2)') 'diag' ,io
   if (io< 10) write (fname,'(a5,i1)') 'diag0',io
   write(99,*)'pe0 writes Energy data on: '//fname//'.dat'

   open(lun,file=fname//'.dat',form='formatted')

   write (lun,*)'nbd_tot  mods   mond_ind'
   write (lun,'(3i8)')ipar(1:3)

   write (lun,*)' totm    tdyn   vir  '
   write (lun,'(3f7.3)')totm,tdyn,vir_0
   envar=22
    write (lun,'(2a6)')'steps',' envar'
    write (lun,'(2i6)')itg,envar
    do k=1,envar
     write (lun,*)varlb(k)
     write (lun,'(6e12.4)')enfun(1:itg,k)
    end do
   close(lun)
    
   if(itnde >0)then
   if (iout<100) write (fname,'(a4,i2)') 'NdEs' ,io
   if (iout< 10) write (fname,'(a5,i1)') 'NdEs0',io
   open(lun,file=fname//'.dat',form='formatted')
   write(99,*)'pe0 writes Energy distributions on: '//fname//'.dat'
    write(lun,*)' NdE  distribution on a grid '
    write(lun,'(2a6)')'steps ','Egrid'
    write(lun,'(2i6)')itnde,ne
    do k=1,itnde
     write(lun,*)'time'
     write(lun,'(e12.4)')entime(k)
     write(lun,*)'Emin,    Emax'
     write (lun,'(2e12.4)')fdes(-1:0,k)
     write(lun,*)'total NdE'
     write (lun,'(5e12.4)')fdes(1:ne,k)
    enddo
   close(lun)
   endif
  endif
!-----------------------
 end subroutine endata
!===================================== 
! Reads the previous output file for a restart
!===================================== 
  subroutine read_en_data(iot,itg1)
   integer,intent(in) :: iot
   integer,intent(out) :: itg1
   integer:: k,nbd1,nvar1,nm,min
   real :: tdyn1,totm1,vir0
   character(6) :: fnme='      '
   integer,parameter ::  lun1=30

   if (iot< 10) write (fnme,'(a5,i1)') 'diag0',iot
!  write(99,*)'reads Energy data on file: '//fnme//'.dat'

   open(lun1,file=fnme//'.dat',form='formatted')

   read(lun1,*)
   read(lun1,'(3i8)')nbd1,nm,min
   read(lun1,*)
   read (lun1,'(3f7.3)')totm1,tdyn1,vir0
   read (lun1,*)
   read(lun1,*)itg1,nvar1
    do k=1,nvar1
     read (lun1,*)
     read (lun1,'(6e12.4)')enfun(1:itg1,k)
    end do
   close(lun1)
 end subroutine read_en_data
!====================================================================
!====================================================================

 end module setio

!------------------------------
!last change 5 sept 2008
!------------------------------
! in subroutine grid_bdata 
!  corrected indices of fdata(1,1,1:nph2)=phi(0:nph2-1)

