!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  

 module util

 contains
!========================
! To LU decompose the square a(0:n,0:n) matrix
! The decomposed matrix is overwritten on am()
! (from the Numerical Recipes library)
!=====================================
 subroutine ludcmp(am,n)
 integer,intent(in) :: n
  real,intent(inout) :: am(0:n,0:n)
  real,parameter::epsa=1.e-10
  integer :: j,i,k,k0
  real:: suma,dum

  do  j=0,n
   if(j >0)then
    do  i=0,j-1
     suma=am(i,j)
     if(i >0)then
      k0=max(0,i-2)
      do k=k0,i-1
       suma=suma-am(i,k)*am(k,j)
      enddo
      am(i,j)=suma
     end if
    end do
   end if
   do i=j,n
    suma=am(i,j)
    if(j >0)then
     k0=max(0,j-2)
     do k=k0,j-1
      suma=suma-am(i,k)*am(k,j)
     end do
     am(i,j)=suma
    end if
   end do
   if(j< n)then
     dum=1./(am(j,j)+epsa)
     do i=j+1,n
      am(i,j)=am(i,j)*dum
     end do
   endif
  end do
  do j=0,n
   am(j,j)=1./(am(j,j)+epsa)
  end do
 end subroutine ludcmp
!============================
! To construct a scalar value plgdr of the P(l,x,m), x=cos(theta) Legendre polynomial
! (from the Numerical Recipes library)
!=============================
 real function plgdr(l,m,x)

  integer,intent(in) :: l,m
  real,intent(in) :: x
  real:: fact,pmm,pmmp1,ag,pll,somx2
  integer :: i,ll

   pmm=1.
   if(m.gt.0)then
    somx2=sqrt((1.0-x)*(1.0+x))
   fact=1.0
   do i=1,m
    pmm=-pmm*fact*somx2
    fact=fact+2.0
   end do
  endif
  if(l.eq.m)then
   plgdr=pmm
  else
   pmmp1=x*float(2*m+1)*pmm
   if(l.eq.m+1)then
    plgdr=pmmp1
   else
    do ll=m+2,l
     ag=float(2*ll-1)
     pll=(x*ag*pmmp1-float(ll+m-1)*pmm)/float(ll-m)
     pmm=pmmp1
     pmmp1=pll
    end do
    plgdr=pll
   end if
  end if
 end function plgdr
end module util
!=================================
! Allocation of the main grid variables defined on the
! full grid. To be used by all Pe's in parallel running
!=================================
module griddata 
 real,allocatable,dimension(:,:,:)::  aux,den,Gr,Gth,Gph
 contains 
 subroutine mainv_alloc(n1,n2,n3,mems)
  integer,intent(in) :: n1,n2,n3
  integer,intent(inout) :: mems

  allocate(aux(0:n1,0:n2-1,0:n3-1))
  allocate(den(-1:n1+1,-1:n2,0:n3))
  allocate(Gr(-1:n1+1,-1:n2,0:n3))
  allocate(Gth(-1:n1+1,-1:n2,0:n3))
  allocate(Gph(-1:n1+1,-1:n2,0:n3))

  aux=0.0
  den=0.0;Gr=0.0
  Gph=0.0; Gth=0.0
  mems=mems+size(aux)+size(den)+size(Gr)+size(Gth)+size(Gph)
 end subroutine mainv_alloc
end module griddata 
!====================================
! Declaration of the grid parameters and arrays
! to be used in the work module for Poisson solvers
!==================================   
module settings
 implicit none
!-------------------
 integer ::  nr,nth,nph,hnph,nth_loc,nph_loc,hnph_loc
 real ::     pi,dx,dth,dth1,dth2,dph,dph1,dx1,dx2,rad_sc
 real ::     dt_iter,asq0,g_0,mass,r_max,tol
 integer :: rmap,rder,spl_ord,ier,itermax,lmax,mond_ind,tol_ind
 integer :: mrate,mem_size,part_size

 real :: tnow,tmax,dt_min,dtime,hdt,cfl,tdyn,dtdyn,dmax,epf,rh,divg_max
 real :: totm,pmass,vir_0,gravc 
   
 real,allocatable,dimension(:):: sth,cth,csd,th,sph,cph,phi
 real,allocatable   ::  r(:,:),drv(:,:),alm(:,:),alm1(:,:)
 real,allocatable,dimension(:,:,:):: thmat,alp,alp_inv
 real,allocatable,dimension(:,:,:):: lpl_rmat,pinv
 real,allocatable,dimension(:,:) :: lfact,dmat0,dmat1,dmat1_inv,dmat

end module settings

module laplclib

 use util 
 use settings
 implicit none

 contains
!===========================
! Sets the (r,theta,phi) grid arrays
!========================================
 subroutine set_rthph_grid

  integer :: i,iy
  real :: xx,dy,sc
  real*16 :: mylfact

  allocate(sth(-1:nth),cth(-1:nth),th(-1:nth),csd(0:nth))
  allocate(sph(0:nph),cph(0:nph),phi(0:nph))
  allocate(r(0:nr,5),drv(0:nr,0:nth-1))
  allocate(lfact(0:nr,0:nth-1))

  pi=acos(-1.0)
  sc=rad_sc

  gravc=1.0
  gravc=4.0*pi*gravc
  dx=0.5*pi/float(nr+1)
  dx1=1./dx
  dx2=dx1*dx1
  r=0.0
!=====================
 select case(rmap)
! The radial grid using the r=sc*tan(x), x=[0,pi/2], mapping
! r(i,1)=r, r(i,2)=r^2, r(i,3)=[dr/dx]^(-1)
! r(i,4),r(i,5) utility arrays for second derivative in Laplace operator 
!==========================
  case(1)
  do i=0,nr
   xx=dx*(float(i)+0.5)
   r(i,1)=sc*tan(xx)
   r(i,2)=r(i,1)*r(i,1)
   r(i,3)=dx*sc/(cos(xx)*cos(xx))   != dr=dx*dr/dx
   r(i,3)=1.0/r(i,3)                ! the inverse 
   r(i,4)=sin(xx)*cos(xx)/dx
   r(i,5)=r(i,3)-dx*r(i,4)*r(i,4)
  end do
!====================
! The radial grid using the r=sc*tan^2(x), x=[0,pi/2], mapping
! r(i,1)=r, r(i,2)=r^2, r(i,3)=[dr/dx]^(-1)
!==========================
  case(2)
  do i=0,nr
   xx=dx*(float(i)+0.5)
   r(i,1)=sc*tan(xx)*tan(xx)
   r(i,2)=r(i,1)*r(i,1)
   r(i,3)=2.0*sc*dx*tan(xx)/(cos(xx)*cos(xx))
   r(i,3)=1.0/r(i,3)
   r(i,4)=r(i,1)*r(i,3)
  end do
 end select
!=====================
! lfact(r,l)=gives the geometrical factor in Poisson formula
! for Legendre components of the Newtonian potential at large distances
! i.e. pot(nr,l,m) at the r=r(nr,1) grid point)
!===========================================
 do i=0,nr
  lfact(i,0)=r(i,1)/r(nr,1)
  do iy=1,lmax
   mylfact = lfact(i,iy-1)
   mylfact = mylfact*lfact(i,0)
   lfact(i,iy) = mylfact
  end do
 end do
 do i=0,nr
  lfact(i,0:lmax)=lfact(i,0:lmax)/r(i,3)
 end do
!============================
! Defines the uniform phi, theta grids
! th-variables are defined on [dth/2,......,pi-dth/2] grid points
! (avoiding the pole singularities)
!=================================  
  hnph=nph/2
  hnph_loc=nph_loc/2
  dph=2.0*pi/float(nph)
  dph1=1.0/dph
  do iy=0,nph
   phi(iy)=dph*iy
   cph(iy)=cos(dph*iy)
   sph(iy)=sin(dph*iy)
  end do

  dth=pi/float(nth)
  dth1=1./dth
  dth2=dth1*dth1
  dy=dth

  do iy=0,nth-1
   csd(iy)=cos(dy*float(iy))-cos(dy*float(iy+1))
  end do
  
  do iy=0,nth
   th(iy)=dy*(float(iy)+0.5)
   sth(iy)=sin(th(iy))
   cth(iy)=cos(th(iy))
  end do
 end subroutine set_rthph_grid
!============================================
! Allocates all the constant arrays needed in Poisson solvers routines:
! alp(th_j,l,m) contains the grid values of P(l,theta_j,m) Legendre polynomials 
! alp_inv(th_j,l,m) contains the grid values of the exact inverse Legebdre polynomials 
!====================================
 subroutine alv_alloc(n1,n2,n3)
  integer,intent(in) :: n1,n2,n3

  allocate(alp(0:n2,0:n2,0:n3/2-1),alp_inv(0:n2,0:n2,0:n3/2-1))
  allocate(thmat(0:n2,1:3,0:1))
  allocate(lpl_rmat(0:n1,0:n2,6),pinv(0:n1,0:n2,1:6))
  allocate(dmat1(0:n1,0:4),dmat0(0:n1,0:4))
  allocate(dmat1_inv(0:n1,0:4))

  mem_size=mem_size+size(alp)+size(alp_inv)+size(lpl_rmat)+size(pinv)

 end subroutine alv_alloc
!===========================
! Constructs the matrix of the coefficients of radial  finite differences 
! derivatives :
! dmat0 discretizes  Gr=d(Pot)/dr
! dmat1 discretizes  rD_r(Gr)+2Gr=r(i,4)DGr+2Gr
!==============================
 subroutine set_rad_mat
  integer ::i

  real,parameter :: b0=-1.5,b1=2.0,b2=-0.5
  dmat0=0.0
  dmat1=0.0
  dmat1_inv=0.0

   i=0
    dmat0(i,2)=b0*r(i,3)
    dmat0(i,3)=b1*r(i,3)
    dmat0(i,4)=b2*r(i,3)
  do i=1,nr-1
   dmat0(i,1)=-0.5*r(i,3)
   dmat0(i,2)=0.0
   dmat0(i,3)=0.5*r(i,3)
  end do
  i=nr
   dmat0(i,1)=-r(i,3)
   dmat0(i,2)=r(i,3)
   dmat0(i,3)=0.0
   dmat1=0.0
   do i=1,nr-1
    dmat1(i,1)=-0.5*r(i,4)
    dmat1(i,2)=2.0
    dmat1(i,3)=0.5*r(i,4)
   end do
   i=0
    dmat1(i,2)=2.0+b0*r(i,4)
    dmat1(i,3)=b1*r(i,4)
    dmat1(i,4)=b2*r(i,4)
   i=nr
    dmat1(i,2)=2.0+r(i,4)
    dmat1(i,1)=-r(i,4)
!---------------------
   dmat1_inv(0:nr,0:4)=dmat1(0:nr,0:4)

   dmat1_inv(0,2)=1./dmat1_inv(0,2)
   do i=1,nr-1
    if(i==2)dmat1_inv(i-1,3)=dmat1_inv(i-1,3)-dmat1_inv(i-1,1)*dmat1_inv(i-2,4)
    dmat1_inv(i,1)=dmat1_inv(i,1)*dmat1_inv(i-1,2)
    dmat1_inv(i,2)=dmat1_inv(i,2)-dmat1_inv(i,1)*dmat1_inv(i-1,3)
    dmat1_inv(i,2)=1.0/dmat1_inv(i,2)
   end do
   i=nr
    dmat1_inv(i,1)=dmat1_inv(i,1)-dmat1_inv(i,0)*dmat1_inv(i-2,3)
    dmat1_inv(i,1)=dmat1_inv(i,1)*dmat1_inv(i-1,2)
    dmat1_inv(i,0)=dmat1_inv(i,0)*dmat1_inv(i-1,2)

    dmat1_inv(i,2)=dmat1_inv(i,2)-dmat1_inv(i,1)*dmat1_inv(i-1,3)
    dmat1_inv(i,2)=1.0/dmat1_inv(i,2)
!------------------------
 end subroutine set_rad_mat
!===================================
! Defines the Legendre polynomial alp(th_j,l,m) and the polynomials
! alp_inv(th_j,l,m) for the discrete Legendre transforms
!================================= 
 subroutine set_leg

  integer:: l,m,s,is,iy,k,k0,j,n
  real::  dy,bb,x,y,anorm,dyk,sumk,sumj,afac2,sumc
  real*16 :: mysqrt, afac1

  allocate(alm(0:nth,0:nth),alm1(0:nth,0:nth))
  alm=0.0
  dth=pi/float(nth)
  dth1=1./dth
  dth2=dth1*dth1
  dy=dth

  do iy=0,nth-1
   csd(iy)=cos(dy*float(iy))-cos(dy*float(iy+1))
  end do
  
  do iy=0,nth
   th(iy)=dy*(float(iy)+0.5)
   sth(iy)=sin(th(iy))
   cth(iy)=cos(th(iy))
  end do
                          !--------- The legenre functions at th_(j+1/2)
  alp=0.0
  is=1
  do m=0,hnph-1    
   do l=m,nth-1
    bb=sqrt(float(2*l+1))
    k=l-m
    afac1=1.0
    if(m >0)then
     do s=1,2*m
      afac1=afac1*float(k+s)
     end do
    endif
    mysqrt=sqrt(afac1)
    bb=bb/mysqrt
    do k=0,nth-1
     x=cth(k)
     alp(k,l,m)=bb*plgdr(l,m,x)
    end do
    alp(nth,l,m)=is*alp(nth-1,l,m)
   end do
   is=-is
  end do
!==========================
!defines the Pade matrix for compact th-derivative
! NOT USED in the present code Version 
!=============================
   do j=0,nth-1
    thmat(j,1,0:1)=1.0/3.0
    thmat(j,2,0:1)=4.0/3.0
    thmat(j,3,0:1)=1.0/3.0
   end do
   thmat(0,1,0:1)=0.0
   thmat(0,2,0)=1.0
   thmat(0,2,1)=5.0/3.0

   thmat(nth-1,3,0:1)=0.0
   thmat(nth-1,2,0)=1.0
   thmat(nth-1,2,1)=5.0/3.0
                          ! set the inverse matrix
   thmat(0,2,0:1)=1./thmat(0,2,0:1)
   do j=1,nth-1
    thmat(j,1,0:1)=thmat(j,1,0:1)*thmat(j-1,2,0:1)
    thmat(j,2,0:1)=thmat(j,2,0:1)-thmat(j,1,0:1)*thmat(j-1,3,0:1)
    thmat(j,2,0:1)=1.0/thmat(j,2,0:1)
   end do
!-------------------------------------------------------------
!    The inverse Legendre transform
!-------------------------------------------------------------
 anorm=2./float(nth)

 do m=0,hnph-2,2   
  do l=m,nth-1
   k0=1
   if(mod(l,2)==0)k0=0
   do  k=k0,l,2
    sumk=.0
    do  j=0,nth-1
     sumk=sumk+alp(j,l,m)*cos(k*th(j))
    end do
    alm(k,l)=anorm*sumk
                         ! the cosine transform of P_{l,m}(cos(th))
   end do
   if(k0==0)alm(0,l)=0.5*alm(0,l)
!
   do n=k0,nth-1,2
    sumc=0.0
    do k=k0,l,2
     afac1=1.0-float((k+n)*(k+n))
     afac2=1.0-float((k-n)*(k-n))
     sumc=sumc+alm(k,l)*(1.0/afac1+1.0/afac2)
    end do
    alm1(n,l)=0.5*sumc
   end do
   alm1(0,l)=0.5*alm1(0,l)

   do j=0,nth-1
    sumj=0.0
    do n=k0,nth-1,2
     sumj=sumj+alm1(n,l)*cos(n*th(j))
    end do
    alp_inv(j,l,m)=anorm*sumj
   end do
  end do
 end do
 alm=0.0
 alm1=0.0
!----------------- odd m modes : sine transform
 do m=1,hnph-1,2   
  do l=m,nth-1
   k0=1
   do  k=k0,l
    sumk=.0
    do  j=0,nth-1
     sumk=sumk+alp(j,l,m)*sin(k*th(j))
    end do
    alm(k,l)=anorm*sumk
   end do
                         ! the cosine transform of P_{l,m}(cos(th))
!
   do n=k0,nth-1
    sumc=0.0
    do k=k0,l
     if(mod((k+n),2)==0)then
      afac1=(1.0-float((k+n)*(k+n)))
      afac1=1.0/afac1
      afac2=(1.0-float((k-n)*(k-n)))
      afac2=1.0/afac2
      sumc=sumc+alm(k,l)*(afac2-afac1)
     endif
    end do
    alm1(n,l)=0.5*sumc
   end do
   do j=0,nth-1
    sumj=0.0
    do n=k0,nth-1
     sumj=sumj+alm1(n,l)*sin(n*th(j))
    end do
    alp_inv(j,l,m)=anorm*sumj
   end do
  end do
 end do
!----------------------
  deallocate(alm,alm1)
                     ! the Dv=r*r*dr*2*alp_inv(0)*dphi spherical volume element
  do j=0,nth-1
   do k=0,nr
    drv(k,j)=r(k,2)*2.0*alp_inv(j,0,0)*dph/r(k,3)
   end do
  end do 

 end subroutine set_leg
!===============================
! To check orthogonality and unit Normalization
!=================================
 subroutine check_orth

  integer :: j,l,l1,m,m1
  real :: sum0
   do m=0,lmax
    write(20,*)'m= ',m
     do l=m,lmax
      do l1=m,lmax
      sum0=0.0
       do j=0,nth-1
       sum0=sum0+alp(j,l,m)*alp_inv(j,l1,m)
       end do
       if(abs(sum0)>0.01)write(20,*)l,l1,sum0
      end do
      end do
      end do
 end subroutine check_orth
!==================================
! Finite differences (or spectral) for first derivatives along the phi coordinate
!====================================
 subroutine Ph_der(u,v,n1,n2,n3,lm,ml,ord)
  integer,intent(in) :: n1,n2,n3,lm,ml,ord
  real,intent(in):: u(0:n1,0:n2-1,0:n3-1)
  real,intent(out):: v(0:n1,0:n2-1,0:n3-1)
  real :: sum0
  integer :: i,j,l,m,m1,m2
  real,parameter :: c1=2.0/3.0,c2=-1.0/12.0
  !real,parameter :: c1=0.5,c2=0.0

  v=0.0
  select case(ord)
  case(0)                    !in (r,th,phi) space central FD
                             ! uses 4th order finite differences
  if(n3<=6)return

   m=0
    v(0:n1,0:n2-1,m)=c1*(u(0:n1,0:n2-1,1)-u(0:n1,0:n2-1,n3-1))+c2*(&
                         u(0:n1,0:n2-1,2)-u(0:n1,0:n2-1,n3-2))
   m=1
    v(0:n1,0:n2-1,m)=c1*(u(0:n1,0:n2-1,2)-u(0:n1,0:n2-1,0))+c2*(&
                         u(0:n1,0:n2-1,3)-u(0:n1,0:n2-1,n3-1))
   do m=2,n3-3
    v(0:n1,0:n2-1,m)=c1*(u(0:n1,0:n2-1,m+1)-u(0:n1,0:n2-1,m-1))+c2*(&
                         u(0:n1,0:n2-1,m+2)-u(0:n1,0:n2-1,m-2))
   end do
   m=n3-2
    v(0:n1,0:n2-1,m)=c1*(u(0:n1,0:n2-1,n3-1)-u(0:n1,0:n2-1,n3-3))+c2*(&
                         u(0:n1,0:n2-1,0)-u(0:n1,0:n2-1,n3-4))
   m=n3-1
    v(0:n1,0:n2-1,m)=c1*(u(0:n1,0:n2-1,0)-u(0:n1,0:n2-1,n3-2))+c2*(&
                         u(0:n1,0:n2-1,1)-u(0:n1,0:n2-1,n3-3))

   v(0:n1,0:n2-1,0:n3-1)=dph1*v(0:n1,0:n2-1,0:n3-1)

  case(1)                     !spectral in (r,th,m) space
  do m=0,ml
   m1=2*m
   m2=m1+1 
   do l=0,n2-1
    v(0:n1,l,m1)=-float(m)*u(0:n1,l,m2)
    v(0:n1,l,m2)=float(m)*u(0:n1,l,m1)
   end do
  end do

  end select
 end subroutine Ph_der
!=======================
! Finite differences for radial derivatives
! Uses coefficients stored in dmat0 or dmat1
!===============================
subroutine Rad_der(u,v,n1,n2,n3,rmod)

  integer,intent(in) :: n1,n2,n3,rmod
  real,intent(in):: u(0:n1,0:n2-1,0:n3-1)
  real,intent(out):: v(0:n1,0:n2-1,0:n3-1)

  integer :: i,j,m,k,m1,m2

  v=0.0
  select case(rmod)
  case(0)
!                         The radial derivative
  do m=0,n3-1
   do j=0,n2-1
    do i=1,n1-1
     v(i,j,m)=dmat0(i,3)*(u(i+1,j,m)-u(i-1,j,m)) 
    end do
    i=0
    v(i,j,m)=dot_product(dmat0(i,2:4),u(i:i+2,j,m))
    i=n1
    v(i,j,m)=dot_product(dmat0(i,1:3),u(i-2:i,j,m))
   end do
  end do
  case(1)
!                         The radial rDiv=rd/dr +2 operator 
  do m=0,n3-1
   do j=0,n2-1
    do i=1,n1-1
     v(i,j,m)=dot_product(dmat1(i,1:3),u(i-1:i+1,j,m)) 
    end do
    i=0
    v(i,j,m)=dot_product(dmat1(i,2:4),u(i:i+2,j,m))
    i=n1
    v(i,j,m)=dot_product(dmat1(i,1:3),u(i-2:i,j,m))
   end do
  end do
 end select

 end subroutine Rad_der
!===========================================
! Discretizes the laplacian [r(d/dr)+2]d/dr-l(l+1)I  operator by 
! matrix multiplication dmat0*dmat1
! The resulting five-diagonal matrix is stored in lpl_rmat(r_i,l,1:5)
! The inverse matrix, to solve the Poisson equation, is stored in pinv(r_i,l,1:5)
!================================================================= 
 subroutine set_poisson_mat(lm)

  integer,intent(in) :: lm
  integer :: l,i,aind,is
!---------------------
  lpl_rmat=0.0
!---------------------------
   i=0
    lpl_rmat(i,0,3)=dmat1(i,2)*dmat0(i,2)+dmat1(i,3)*dmat0(i+1,1)
    lpl_rmat(i,0,4)=dmat1(i,2)*dmat0(i,3)+dmat1(i,3)*dmat0(i+1,2)+ &
                dmat1(i,4)*dmat0(i+2,1)
    lpl_rmat(i,0,5)=dmat1(i,2)*dmat0(i,4)+dmat1(i,3)*dmat0(i+1,3)+ &
                dmat1(i,4)*dmat0(i+2,2)
    lpl_rmat(i,0,6)=dmat1(i,4)*dmat0(i+2,3)
!-------------
   i=1
    lpl_rmat(i,0,1)=0.0
    lpl_rmat(i,0,5)=dmat1(i,3)*dmat0(i+1,3)
    lpl_rmat(i,0,2)=dmat1(i,1)*dmat0(i-1,2)+dmat1(i,2)*dmat0(i,1)
    lpl_rmat(i,0,4)=dmat1(i,2)*dmat0(i,3)+dmat1(i,3)*dmat0(i+1,2)+ &
                dmat1(i,1)*dmat0(i-1,4)
    lpl_rmat(i,0,3)=dmat1(i,1)*dmat0(i-1,3)+ dmat1(i,2)*dmat0(i,2)+&
                dmat1(i,3)*dmat0(i+1,1)
   do i=2,nr-1
    lpl_rmat(i,0,1)=dmat1(i,1)*dmat0(i-1,1)
    lpl_rmat(i,0,5)=dmat1(i,3)*dmat0(i+1,3)
    lpl_rmat(i,0,2)=dmat1(i,1)*dmat0(i-1,2)+dmat1(i,2)*dmat0(i,1)
    lpl_rmat(i,0,4)=dmat1(i,2)*dmat0(i,3)+dmat1(i,3)*dmat0(i+1,2)
    lpl_rmat(i,0,3)=dmat1(i,1)*dmat0(i-1,3)+ dmat1(i,2)*dmat0(i,2)+&
                dmat1(i,3)*dmat0(i+1,1)
   end do
   i=nr-1
   lpl_rmat(i,0,5)=0.0
   i=nr
    lpl_rmat(i,0,1)=dmat1(i,1)*dmat0(i-1,1)
    lpl_rmat(i,0,2)=dmat1(i,2)*dmat0(i,1)+dmat1(i,1)*dmat0(i-1,2)
    lpl_rmat(i,0,3)=dmat1(i,1)*dmat0(i-1,3)+dmat1(i,3)*dmat0(i,2)
!-------------
   do l=0,lm 
    aind=l*(l+1)
    lpl_rmat(0:nr,l,1:6)=lpl_rmat(0:nr,0,1:6)
    do i=0,nr
     lpl_rmat(i,l,3)=lpl_rmat(i,0,3)-aind/r(i,1)
    end do
   end do
! --------------------------
   call Amsix(lm)

 end subroutine set_poisson_mat
!++++++++++++++++++++++++++++++++++++++++
! LU decomposition of the Laplacian matrix to compute the inverse Laplacian
!=========================================
 subroutine Amsix(lm)
   integer,intent(in) :: lm
   integer :: i,l,nb

   nb=nr-1
   allocate(alm(0:nb,0:nb))
   do l=0,lm
    alm=0.0
    alm(0,0:3)=lpl_rmat(0,l,3:6)
    alm(1,0:3)=lpl_rmat(1,l,2:5)
    do i=2,nb-2
     alm(i,i-2:i+2)=lpl_rmat(i,l,1:5)
    end do
    i=nb-1
    alm(i,i-2:i+1)=lpl_rmat(i,l,1:4)
    i=nb
    alm(i,i-2:i)=lpl_rmat(i,l,1:3)
    call ludcmp(alm,nb)

    pinv(0,l,3:6)=alm(0,0:3)
    pinv(1,l,2:5)=alm(1,0:3)
    do i=2,nb-2
     pinv(i,l,1:5)=alm(i,i-2:i+2)
    end do
    i=nb-1
    pinv(i,l,1:4)=alm(i,i-2:i+1)
    i=nb
    pinv(i,l,1:3)=alm(i,i-2:i)
    pinv(i,l,4)=lpl_rmat(i,l,4)
   end do
  if(allocated(alm))deallocate(alm)
 end subroutine Amsix

 end module laplclib
!=========================================
!
 module work

  use fft_lib
  use parallel
  use griddata
  use laplclib
  implicit none

   real,allocatable,dimension(:,:,:):: mu,aux1,den1
   real,allocatable,dimension(:,:,:):: dGr,dGth,dGph
   integer,parameter :: MAXV=2, MINV=1, SUMV=0 
  contains
!==============================
! Allocates all arrays needed for MOND and Newtonian solvers
! Here n1=Nr, n2=Nth, n2_loc=Nth/NPE for parallel computations,n3=Nph
!========================================
 subroutine v_alloc(n1,n2,n2_loc,n3)

  integer,intent(in) :: n1,n2,n2_loc,n3

  allocate(mu(0:n1,0:n2_loc-1,0:n3-1))
  allocate(den1(0:n1,0:n2_loc-1,0:n3-1))
  allocate(dGr(0:n1,0:n2_loc-1,0:n3-1))
  allocate(dGth(0:n1,0:n2_loc-1,0:n3-1))
  allocate(dGph(0:n1,0:n2_loc-1,0:n3-1))
  allocate(lbdth(n1+1,n3,2),rbdth(n1+1,n3,2))
  mem_size=0
  mu=1.0
  den1=0.0;dGr=0.0
  dGph=0.0; dGth=0.0
  lbdth=0.0
  rbdth=0.0
  mem_size=mem_size+5*size(mu)+2*size(lbdth)
  allocate(aux1(0:n1,0:n2-1,0:n3-1))
  aux1=0.0
  mem_size=mem_size+size(aux1)

 end subroutine v_alloc
!===========================
 subroutine all_dealloc
 
  if(allocated(mu))deallocate(mu)
  if(allocated(den1))deallocate(den1)
  if(allocated(dGr))deallocate(dGr)
  if(allocated(dGth))deallocate(dGth)
  if(allocated(dGph))deallocate(dGph)
  if(allocated(lbdth))deallocate(lbdth)
  if(allocated(rbdth))deallocate(rbdth)

  if(allocated(den))deallocate(den)
  if(allocated(aux))deallocate(aux)
  if(allocated(aux1))deallocate(aux1)
  if(allocated(Gth))deallocate(Gr)
  if(allocated(Gth))deallocate(Gth)
  if(allocated(Gph))deallocate(Gph)

 end subroutine all_dealloc
!===================================
! Finite differences for theta-derivatives
! Executed in parallel, for each n2=Nth/NPE angular sector
!==================================== 
 subroutine Th_der(u,v,n1,n2,n3,lpe,is)

  integer,intent(in) :: n1,n2,n3,is,lpe
  real,intent(in):: u(0:n1,0:n2-1,0:n3-1)
  real,intent(out):: v(0:n1,0:n2-1,0:n3-1)
  integer :: i,j,m1,k,mh
  real,parameter :: c1=2.0/3.0,c2=-1./12.
  !real,parameter :: c1=0.5,c2=0.0

                             ! uses central 4th order finite differences
  v=0.0
  
  mh=n3/2
  if(prl)then
   call exchange_thbd_data(u,n1,n2,n3,2)
   do k=0,n3-1
    m1=k+mh
    if(m1 >n3-1)m1=m1-n3
    if(lpe==0)then
     do j=1,2
      lbdth(1:n1+1,k+1,j)=is*u(0:n1,j-1,m1)
     end do
    endif
    if(lpe==pe_max)then
     do j=1,2
      rbdth(1:n1+1,k+1,j)=is*u(0:n1,n2-j,m1)
     end do
    endif
    v(0:n1,0,k)=c1*(u(0:n1,1,k)-lbdth(1:n1+1,k+1,1)) +c2*( &
                   u(0:n1,2,k)-lbdth(1:n1+1,k+1,2))
    v(0:n1,1,k)=c1*(u(0:n1,2,k)-u(0:n1,0,k)) +c2*( &
                   u(0:n1,3,k)-lbdth(1:n1+1,k+1,1))
    v(0:n1,n2-1,k)=c1*(rbdth(1:n1+1,k+1,1)-u(0:n1,n2-2,k))+c2*( &
                      rbdth(1:n1+1,k+1,2)-u(0:n1,n2-3,k))
    v(0:n1,n2-2,k)=c1*(u(0:n1,n2-1,k)-u(0:n1,n2-3,k))+c2*( &
                      rbdth(1:n1+1,k+1,1)-u(0:n1,n2-4,k))
   enddo
  else
   do k=0,n3-1
    m1=k+mh
    if(m1 >n3-1)m1=m1-n3
    v(0:n1,0,k)=c1*(u(0:n1,1,k)-is*u(0:n1,0,m1))+c2*( &
                   u(0:n1,2,k)-is*u(0:n1,1,m1))
    v(0:n1,1,k)=c1*(u(0:n1,2,k)-u(0:n1,0,k))+c2*( &
                   u(0:n1,3,k)-is*u(0:n1,0,m1))
    v(0:n1,n2-1,k)=c1*(is*u(0:n1,n2-1,m1)-u(0:n1,n2-2,k))+c2*( &
                       is*u(0:n1,n2-2,m1)-u(0:n1,n2-3,k))
    v(0:n1,n2-2,k)=c1*(u(0:n1,n2-1,k)-u(0:n1,n2-3,k))+c2*( &
                      is*u(0:n1,n2-1,m1)-u(0:n1,n2-4,k))
   end do 
  endif 
  do k=0,n3-1
   do j=2,n2-3
    v(0:n1,j,k)=c1*(u(0:n1,j+1,k)-u(0:n1,j-1,k))+c2*( &
                    u(0:n1,j+2,k)-u(0:n1,j-2,k))
   end do
   v(0:n1,0:n2-1,k)=dth1*v(0:n1,0:n2-1,k)
  end do
 end subroutine Th_der
!=================================
 subroutine check_rder
  integer :: i
  real :: r2
   do i=0,nr
   r2=r(i,1)*r(i,1)
   dGr(i,0:nth-1,0:nph-1)=r(i,1)/(1.+r2)
   end do
   call Rad_der(dGr,den1,nr,nth,nph,1)
   write(29,*)'R_der',nr,nth,nph
  do i=0,nr
   r2=r(i,1)*r(i,1)
   write(29,*)den1(i,10,10),r(i,1)*(3.+r2)/(1.+r2)**2
  end do
 end subroutine check_rder
!=============================
 subroutine check_thder

  integer :: j
   do j=0,nth-1
   dGth(0:nr,j,0:nph-1)=sth(j)*sth(j)
   end do
   call Th_der(dGth,den1,nr,nth_loc,nph,0,1)
  do j=0,nth-1
   write(29,*)dGth(0,j,0),den1(0,j,0),2.*sth(j)*cth(j)
  end do
 end subroutine check_thder
!======================================
! Direct (is=-1) or inverse (is=1) Legendre transforms
! for variables defined on the full grid (Nr,Nth,Nph)
!================================================ 
 subroutine full_lgd_trsf(lm,ml,is)
   integer,intent(in) :: lm,ml,is

   integer :: i,j,l,m,m1,m2
   
   select case(is)
   case(-1)             ! from f(r,theta,m) => f(r,l,m)
   aux1=0.0
   do m=0,min(ml,hnph-1) 
    m1=2*m
    m2=m1+1
    do i=0,nr
     do l=m,lm
      do j=0,nth-1
       aux1(i,l,m1)=aux1(i,l,m1)+aux(i,j,m1)*alp_inv(j,l,m)
       aux1(i,l,m2)=aux1(i,l,m2)+aux(i,j,m2)*alp_inv(j,l,m)
      end do
     end do
    end do
   end do
   aux(0:nr,0:nth-1,0:nph-1)=aux1(0:nr,0:nth-1,0:nph-1)
   case(1)             ! from legendre f(r,l,m) to grid space f(r,theta,m)
   aux1=0.0
   do m=0,min(ml,hnph-1)
    m1=2*m
    m2=m1+1
    do i=0,nr
     do j=0,nth-1
      do l=m,lm
       aux1(i,j,m1)=aux1(i,j,m1)+aux(i,l,m1)*alp(j,l,m)
       aux1(i,j,m2)=aux1(i,j,m2)+aux(i,l,m2)*alp(j,l,m)
      end do
     end do
    end do
   end do
   aux(0:nr,0:nth-1,0:nph-1)=aux1(0:nr,0:nth-1,0:nph-1)
   end select
 end subroutine full_lgd_trsf
!======================================
! Direct (is=-1) or inverse (is=1) Legendre transforms
! for variables defined on the partial (th-decomposed) grid (Nr,Nth_loc,Nph)
! Executed in parallel
!================================================ 
 subroutine lgd_trsf(lm,ml,is,pe_loc)
   integer,intent(in) :: lm,ml,is,pe_loc

   integer :: i,j,j1,l,m,m1,m2
   
   select case(is)
   case(-1)             ! from f(r,theta,m) => f(r,l,m)
   aux=0.0
   do m=0,min(ml,hnph-1) 
    m1=2*m
    m2=m1+1
    do l=m,lm
     do j=0,nth_loc-1
      j1=j+nth_loc*pe_loc
      do i=0,nr
       aux(i,l,m1)=aux(i,l,m1)+den1(i,j,m1)*alp_inv(j1,l,m)
       aux(i,l,m2)=aux(i,l,m2)+den1(i,j,m2)*alp_inv(j1,l,m)
      end do
     end do
    end do
   end do
   case(1)             ! from legendre f(r,l,m) to grid space f(r,theta,m)
   den1=0.0   
   do m=0,min(ml,hnph-1)
    m1=2*m
    m2=m1+1
    do j=0,nth_loc-1
     j1=j+nth_loc*pe_loc
     do l=m,lm
      do i=0,nr
       den1(i,j,m1)=den1(i,j,m1)+aux(i,l,m1)*alp(j1,l,m)
       den1(i,j,m2)=den1(i,j,m2)+aux(i,l,m2)*alp(j1,l,m)
      end do
     end do
    end do
   end do
   end select
 end subroutine lgd_trsf
!======================================
! Filtering of density data by mode truncation in Legendre-Fourier space
!=====================================
 subroutine filt(lm)
  integer,intent(in) :: lm
  integer :: i,ml 
                    ! filters vden array
  ml=min(hnph-1,lm)
  call ft1_ph(aux,nr,nth,nph,-1)
  call full_lgd_trsf(lm,ml,-1) ! aux => aux
   dmax=0.0
   do i=0,nr/2
    dmax=max(dmax,abs(aux(i,0,0)))
   end do

  call full_lgd_trsf(lm,ml,1)  ! aux => aux
  call ft1_ph(aux,nr,nth,nph,1)

 end subroutine filt
!===================================
! Solves a Poisson equation in Fourier-Legendre components
! using the stored pinv(:.:.:) coefficients matrix.
! In aux(r_i,l,m) enters the r.h.s. source term; on output the same 
! array contains the resulting potential function
!================================================
 subroutine poisson(nb,lm,ml)
  integer,intent(in) :: nb,lm,ml
  integer :: i,l,m,m1,m2,k 

  do k=0,ml 
   m1=2*k
   m2=m1+1
   do m=m1,m2
   do l=k,lm
    aux(nr-1,l,m)=aux(nr-1,l,m)-pinv(nr-1,l,4)*aux(nr,l,m)
   end do
   do l=k,lm
    aux(1,l,m)=aux(1,l,m)-pinv(1,l,2)*aux(0,l,m)
    do i=2,nb
     aux(i,l,m)=aux(i,l,m)-pinv(i,l,2)*aux(i-1,l,m)- &
                             pinv(i,l,1)*aux(i-2,l,m)
    end do
    aux(nb,l,m)=aux(nb,l,m)*pinv(nb,l,3)
    aux(nb-1,l,m)=aux(nb-1,l,m)- &
                   pinv(nb-1,l,4)*aux(nb,l,m)
    aux(nb-1,l,m)=aux(nb-1,l,m)*pinv(nb-1,l,3)
    do i=nb-2,0,-1
     aux(i,l,m)=aux(i,l,m)-pinv(i,l,4)*aux(i+1,l,m)- &
                             pinv(i,l,5)*aux(i+2,l,m)
     if(i==0)aux(i,l,m)=aux(i,l,m)-pinv(i,l,6)*aux(i+3,l,m)
     aux(i,l,m)=aux(i,l,m)*pinv(i,l,3)
    end do
   end do
   end do
  end do
 end subroutine poisson 
!=============================================
!        Section devoted to Newtonian gravity 
!=============================================
! Sets the far-field boundary conditions and solves for the
! Newtonian potential
!=======================================
 subroutine newt_pot_inv(lmx,ml,nit)
  integer,intent(in) :: lmx,ml,nit
  integer :: i,m,m1,l,k,nbd
  real :: sum0(2)
                !enters aux=den => exit aux=pot in (r,l,m) space) 
  nbd=nr-1
  if(nit==1)then
   do m=0,ml
    m1=2*m
    do l=m,lmx
     sum0=0.0 
     do i=0,nr
      sum0(1)=sum0(1)-lfact(i,l)*aux(i,l,m1)
      sum0(2)=sum0(2)-lfact(i,l)*aux(i,l,m1+1)
     end do
     aux(nr,l,m1)=sum0(1)
     aux(nr,l,m1+1)=sum0(2)
    end do
   end do
  else
   aux(nr,:,:)=0.0
  endif
  call poisson(nbd,lmx,ml)

end subroutine newt_pot_inv
!==========================
! For a provisional potential solution, computes  
! the Laplacian of the potential in Fourier-Legendre components
!======================================
subroutine div_gradp(lm,ml,loc_pe)

  integer,intent(in) :: lm,ml,loc_pe
  integer :: i,l,k,m,m1,m2
                           !enters pot 
                           !exit r*Lapl(pot) in dGr
!----------------------------
  do k=0,ml
   m1=2*k
   m2=m1+1
   do m=m1,m2
    do l=k,lm
    i=0
     den1(i,0,0)=dot_product(lpl_rmat(i,l,3:6),Gr(i:i+3,l,m))
    i=1
     den1(i,0,0)=dot_product(lpl_rmat(i,l,2:5),Gr(i-1:i+2,l,m))
     do i=2,nr-2
      den1(i,0,0)=dot_product(lpl_rmat(i,l,1:5),Gr(i-2:i+2,l,m))
     end do
     i=nr-1
     den1(i,0,0)=dot_product(lpl_rmat(i,l,1:4),Gr(i-2:i+1,l,m))
     i=nr
     den1(i,0,0)=dot_product(lpl_rmat(i,l,1:3),Gr(i-2:i,l,m))
     do i=0,nr
      aux(i,l,m)=den1(i,0,0)
     end do
    end do
   end do
  end do 

 end subroutine div_gradp
!==========================
! Computes the (r,theta,phi) derivatives of the potential
! on the full-grid data array aux(:,:,:) 
!======================================
 subroutine full_gradP
  integer :: i,j,m,m1,mh
  real,parameter :: c1=2.0/3.0,c2=-1.0/12.0
  
  do m=0,nph-1
   do j=0,nth-1
    do i=1,nr-1
     Gr(i,j,m)=dmat0(i,3)*(aux(i+1,j,m)-aux(i-1,j,m)) 
    end do
    i=0
    Gr(i,j,m)=dot_product(dmat0(i,2:4),aux(i:i+2,j,m))
    i=nr
    Gr(i,j,m)=dot_product(dmat0(i,1:3),aux(i-2:i,j,m))
   end do
  end do
  mh=nph/2
  do m=0,nph-1
   m1=m+mh
   if(m1 >nph-1)m1=m1-nph
    Gth(0:nr,0,m)=c1*(aux(0:nr,1,m)-aux(0:nr,0,m1))+c2*( &
                   aux(0:nr,2,m)-aux(0:nr,1,m1))
    Gth(0:nr,1,m)=c1*(aux(0:nr,2,m)-aux(0:nr,0,m))+c2*( &
                     aux(0:nr,3,m)-aux(0:nr,0,m1))
    Gth(0:nr,nth-1,m)=c1*(aux(0:nr,nth-1,m1)-aux(0:nr,nth-2,m))+c2*( &
                      aux(0:nr,nth-2,m1)-aux(0:nr,nth-3,m))
    Gth(0:nr,nth-2,m)=c1*(aux(0:nr,nth-1,m)-aux(0:nr,nth-3,m))+c2*( &
                      aux(0:nr,nth-1,m1)-aux(0:nr,nth-4,m))
   do j=2,nth-3
    Gth(0:nr,j,m)=c1*(aux(0:nr,j+1,m)-aux(0:nr,j-1,m)) +c2*( &
                      aux(0:nr,j+2,m)-aux(0:nr,j-2,m))
   end do
   Gth(0:nr,0:nth-1,m)=dth1*Gth(0:nr,0:nth-1,m)
  end do
  if(nph<=6)return

   m=0
   Gph(0:nr,0:nth-1,m)=c1*(aux(0:nr,0:nth-1,1)-aux(0:nr,0:nth-1,nph-1))+c2*(&
                         aux(0:nr,0:nth-1,2)-aux(0:nr,0:nth-1,nph-2))
   m=1
   Gph(0:nr,0:nth-1,m)=c1*(aux(0:nr,0:nth-1,2)-aux(0:nr,0:nth-1,0))+c2*(&
                         aux(0:nr,0:nth-1,3)-aux(0:nr,0:nth-1,nph-1))
   do m=2,nph-3
    Gph(0:nr,0:nth-1,m)=c1*(aux(0:nr,0:nth-1,m+1)-aux(0:nr,0:nth-1,m-1))+c2*(&
                         aux(0:nr,0:nth-1,m+2)-aux(0:nr,0:nth-1,m-2))
   end do
   m=nph-2
    Gph(0:nr,0:nth-1,m)=c1*(aux(0:nr,0:nth-1,m+1)-aux(0:nr,0:nth-1,m-1))+c2*(&
                         aux(0:nr,0:nth-1,0)-aux(0:nr,0:nth-1,m-2))
   m=nph-1
    Gph(0:nr,0:nth-1,m)=c1*(aux(0:nr,0:nth-1,0)-aux(0:nr,0:nth-1,m-1))+c2*(&
                         aux(0:nr,0:nth-1,1)-aux(0:nr,0:nth-1,m-2))

   Gph(0:nr,0:nth-1,0:nph-1)=dph1*Gph(0:nr,0:nth-1,0:nph-1)

 end subroutine full_gradP
!==================================================
! Solves for Newtonian potential and acceleration
!====================================================
 subroutine newt_acc(loc_pe,loc_iter)
  integer,intent(in) :: loc_pe,loc_iter 

  integer :: itr,i,j,k,j1,l,ml_max
  real,parameter :: small=1.e-08
  real :: rms,rms1,err_max,err1_max
  logical,parameter :: bcst=.true.
!---------------------
!    Computes Newtonian acceleraions via the Poisson equation
!    and potential differentiation  using
!    centered second order FD in all coordinates.
!-----------------------------------------------------------------

  itr=0
  tol=tol_ind*1.e-04
  ml_max=min(lmax,nph/2-1)
  dt_iter=0.9
!=============================
  divg_max=gravc*dmax
!========================
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    den1(0:nr,j,k)=-den(0:nr,j1,k)
   end do
  end do
  call ft1_ph(den1,nr,nth_loc,nph,-1)
  call lgd_trsf(lmax,ml_max,-1,loc_pe) ! exit aux(i,l,m)= local den1 source term

   Gr=0.0              !collect the potential in Gr 
   do k=0,ml_max
    j1=2*k
    do l=k,lmax
     Gth(0:nr,l,j1)=-aux(0:nr,l,j1)
     Gth(0:nr,l,j1+1)=-aux(0:nr,l,j1+1)    !stores density in (l,m) space
    end do
   end do
  do                  !
   itr=itr+1
   call newt_pot_inv(lmax,ml_max,itr) !solves r*&Lapl(phi)=-4*Pi*G*r*rho
   do k=0,ml_max
    j1=2*k
    do l=k,lmax
     Gr(0:nr,l,j1)=Gr(0:nr,l,j1)+ aux(0:nr,l,j1)
     Gr(0:nr,l,j1+1)=Gr(0:nr,l,j1+1)+ aux(0:nr,l,j1+1)   !the new potential
    end do
   end do
   call div_gradp(lmax,ml_max,loc_pe)! enters the pot increment exit aux=Laplace(pot) 
   do k=0,ml_max
    j1=2*k
    do l=k,lmax
     aux(0:nr,l,j1)=Gth(0:nr,l,j1)+ aux(0:nr,l,j1)
     aux(0:nr,l,j1+1)=Gth(0:nr,l,j1+1)+ aux(0:nr,l,j1+1)
    end do
   end do
   err_max=maxval(abs(aux(0:nr-1,0:lmax,0:nph-1)))
   if(prl)then
    call sumreduce_grdata(bcst,aux,aux1,&
                         nr+1,nth,nph)
   err_max=maxval(abs(aux1(0:nr-1,0:lmax,0:nph-1)))
   endif 
                 ! a new source term

   aux(0:nr,0:lmax,0:nph-1)=-dt_iter*aux(0:nr,0:lmax,0:nph-1)
                                !(-(den+r*div(G)) the new source term
           !exit den1=(r*divG+den)
   if(itr >=itermax)exit
   if(err_max<=tol)exit
  enddo
                           ! end of iteration loop
  if(mod(loc_iter,mrate)==0)then
   if(loc_pe==0)then
    write(99,*)'number of loc iterations',itr
    write(99,'(a11,e11.4)')' div_max   ',err_max
   endif
  endif
  do k=0,ml_max
   j1=2*k
   do l=k,lmax
    aux(0:nr,l,j1)=Gr(0:nr,l,j1)
    aux(0:nr,l,j1+1)=Gr(0:nr,l,j1+1)  !the  potential
    end do
   end do
   if(prl)then
    call sumreduce_grdata(bcst,aux,aux1,&
                         nr+1,nth,nph)
    do k=0,ml_max
     j1=2*k
     do l=k,lmax
      aux(0:nr,l,j1)=aux1(0:nr,l,j1)
      aux(0:nr,l,j1+1)=aux1(0:nr,l,j1+1)
     end do
    end do
  endif 
!+++++++++++++++++++++++++++++++++++++
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    den1(0:nr,j,k)=-den(0:nr,j1,k)
   end do
  end do
  call full_lgd_trsf(lmax,ml_max,1)  ! aux =>aux 
  call ft1_ph(aux,nr,nth,nph,1)

  call full_gradP 

  call set_fbd(1)
!+++++++++++++++++++++++++++++++++
! Exit (Gr,Gth,Gph)=(dPot/dr, dPor/dth, dPot/dphi) and aux()=Pot
! Density source den()=4Pi*r*rho is preserved
!-----------------
 end subroutine newt_acc
!++++++++++++++++++++++++++++++++++++++++++++++++
 subroutine Check_Lapl(loc_pe)
  integer,intent(in) :: loc_pe 

  integer :: i,j,k,j1,l,ml_max

  ml_max=min(lmax,nph/2-1)
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    den1(0:nr,j,k)=den(0:nr,j1,k)
   end do
  end do
  call ft1_ph(den1,nr,nth_loc,nph,-1)
  call lgd_trsf(lmax,ml_max,-1,loc_pe)        ! exit aux(i,l,m)= local source term

  do k=0,ml_max
   j1=2*k
   do l=k,lmax
    Gr(0:nr,l,j1)=aux(0:nr,l,j1)
    Gr(0:nr,l,j1+1)=aux(0:nr,l,j1+1)  
   end do
  end do
  call div_gradp(lmax,ml_max,loc_pe) !exit aux= 
  call lgd_trsf(lmax,ml_max,1,loc_pe)        ! exit aux(i,l,m)= local source term
  call ft1_ph(den1,nr,nth_loc,nph,1)

 end subroutine Check_Lapl
!+++++++++++++++++++++++++++++++++
!     Section devoted to Mond Gravity
!+++++++++++++++++++++++++++++++++++++++++
 subroutine mond_init(loc_pe)
  integer,intent(in) :: loc_pe
  integer :: i,j,j1,k,lm
  real :: bb,sf,asq04
  logical,parameter :: bst=.true.

          !initializes the Mond iterations with a sferical symmetric
          !(rot-free) solution mu*Gr=Fr

   asq04=4.0*asq0
   lm=0
   do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    den1(0:nr,j,k)=-den(0:nr,j1,k)
   end do
  end do
  call ft1_ph(den1,nr,nth_loc,nph,-1)
  call lgd_trsf(lm,lm,-1,loc_pe) 
  call newt_pot_inv(lm,lm,1) 
   if(prl)then
    call sumreduce_grdata(bst,aux,aux1,&
                         nr+1,nth,nph)
    aux(0:nr,0,0)=aux1(0:nr,0,0)
   endif 
  call lgd_trsf(lm,lm,1,loc_pe)        ! the potential in the grid space 
  call ft1_ph(den1,nr,nth_loc,nph,1)

  call Rad_der(den1,dGth,nr,nth_loc,nph,0)

   if(mond_ind==2)then
    do k=0,nph-1
     do j=0,nth_loc-1
      do i=0,nr
       bb=dGth(i,j,k)**2
       sf=dGth(i,j,k)/sqrt(bb)
       dGr(i,j,k)=sf*sqrt(bb)
      end do
     end do
    end do
   endif
   if(mond_ind==1)then 
    do k=0,nph-1
     do j=0,nth_loc-1
      do i=0,nr
       bb=dGth(i,j,k)**2
       sf=dGth(i,j,k)/sqrt(bb)
       dGr(i,j,k)=0.5*sf*sqrt(bb+sqrt(bb*(bb+asq04)))
      end do
     end do
    end do
   endif
    dGth=0.0
    dGph=0.0
                     !Exit the Mond solution for angle-averaged density
 end subroutine mond_init
!+++++++++++++++++++++++++++++++
 subroutine psolv(dt,loc_pe,lmx,ml)
   real,intent(in) :: dt 
   integer,intent(in) :: loc_pe,lmx,ml
   integer :: i,j,j1,k
  logical,parameter :: bst=.true.
             ! solves Poisson equation for 
             ! for the potential for all Legendre l modes
             ! enters the source term dGr=-r*(den+div(mu*G))
             ! solves the Poisson equation for the potential increments 
!----------------------------
 do k=0,nph-1
  do j=0,nth_loc-1
   do i=0,nr
    den1(i,j,k)=dt*dGr(i,j,k)/mu(i,j,k) !the source term
   end do
  end do
 end do 
 call ft1_ph(den1,nr,nth_loc,nph,-1)
 call lgd_trsf(lmx,ml,-1,loc_pe)
  aux(nr,:,:)=0.0
  call poisson(nr-1,lmx,ml)
  if(prl)then
   call sumreduce_grdata(bst,aux,aux1,&
                         nr+1,nth,nph)
   do k=0,ml
    j1=2*k
    do j=k,lmx
     aux(0:nr,j,j1)=aux1(0:nr,j,j1)
     aux(0:nr,j,j1+1)=aux1(0:nr,j,j1+1)
    end do
   end do
  endif 
!-------------------
  call lgd_trsf(lmx,ml,1,loc_pe)
  call ft1_ph(den1,nr,nth_loc,nph,1)
                              ! exit pot in den1
!------------------------- explicit centered FD in radius
  call Rad_der(den1,dGr,nr,nth_loc,nph,0)
  call Th_der(den1,dGth,nr,nth_loc,nph,loc_pe,1)
  call Ph_der(den1,dGph,nr,nth_loc,nph,lmx,ml,0)

 do k=0,nph-1
  do j=0,nth_loc-1
   j1=j+loc_pe*nth_loc
   do i=0,nr
    dGth(i,j,k)=dGth(i,j,k)/r(i,1)
    dGph(i,j,k)=dGph(i,j,k)/(sth(j1)*r(i,1))
   end do
  end do
 end do 
!--------------------
 end subroutine psolv
!========================================
 subroutine divG(lpe)
  integer,intent(in) :: lpe 
  integer :: i,j,k,j1,ml

  ml=min(lmax,hnph-1)
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+nth_loc*lpe
    dGth(0:nr,j,k)=sth(j1)*dGth(0:nr,j,k)
   end do
  end do
  
  call Th_der(dGth,den1,nr,nth_loc,nph,lpe,1)

  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+nth_loc*lpe
    dGth(0:nr,j,k)=den1(0:nr,j,k)/sth(j1)
   end do
  end do 

  call Ph_der(dGph,den1,nr,nth_loc,nph,lmax,ml,0)
!----------------------------
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+nth_loc*lpe
    dGth(0:nr,j,k)=dGth(0:nr,j,k)+den1(0:nr,j,k)/sth(j1)
   end do
  end do 

  call Rad_der(dGr,den1,nr,nth_loc,nph,1)  !computes rDGr/Dr +2Gr

  do k=0,nph-1
   do j=0,nth_loc-1 
    do i=0,nr
     dGr(i,j,k)=den1(i,j,k)+dGth(i,j,k)
    end do
   end do
  end do 
!----------------------
 end subroutine divG 

 subroutine Gmond(pe_loc,ig)
  integer,intent(in) :: pe_loc,ig

  integer :: i,j,k,j1
  real :: x

  mu=1.0

  select case(ig)
  case(0)

  case(1)

  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+pe_loc*nth_loc
    do i=0,nr
     x=Gph(i,j1,k)*Gph(i,j1,k)+   &
       Gth(i,j1,k)*Gth(i,j1,k)+   &
       Gr(i,j1,k)*Gr(i,j1,k)
     x=x/asq0
     if(x >0.0)mu(i,j,k)=sqrt(x/(1.0+x))
    end do
   end do
  end do 
  case(2)
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+pe_loc*nth_loc
    do i=0,nr
     x=Gph(i,j1,k)*Gph(i,j1,k)+   &
       Gth(i,j1,k)*Gth(i,j1,k)+   &
       Gr(i,j1,k)*Gr(i,j1,k)
     if(x >0.0)mu(i,j,k)=sqrt(x)
    end do
   end do
  end do 
!-----------------
  end select
 end subroutine Gmond
!----------------
 subroutine div_max(loc_pe)

  integer,intent(in) :: loc_pe
  integer :: i,j,j1,k
  real :: max_dg

   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      dGr(i,j,k)=Gr(i,j1,k)
      dGth(i,j,k)=Gth(i,j1,k)
      dGph(i,j,k)=Gph(i,j1,k)
     end do
    end do
   end do  

   call divG(loc_pe) ! exit dGr=r*div(mu*G)
   max_dg=0.0
   do k=0,nph-1
    do j=0,nth_loc-1
     do i=0,nr
     max_dg=max(max_dg,abs(dGr(i,j,k)/r(i,1))) 
    end do
    end do
   end do
  call allreduce_real_sdata(max_dg,divg_max,MAXV)

 end subroutine div_max

 subroutine div_err(md,loc_pe,rmx,rms,mxind)

  integer,intent(in) :: md,loc_pe
  real,intent(out):: rmx,rms
  integer,intent(out) :: mxind(3)
  integer :: i,j,j1,k

  if(md==1)then

   call Gmond(loc_pe,mond_ind)

   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      dGr(i,j,k)=Gr(i,j1,k)*mu(i,j,k)
      dGth(i,j,k)=Gth(i,j1,k)*mu(i,j,k)
      dGph(i,j,k)=Gph(i,j1,k)*mu(i,j,k)
     end do
    end do
   end do  

   call divG(loc_pe) ! exit dGr=r*div(mu*G)

   do k=0,nph-1
    do j=0,nth_loc-1
     j1=j+loc_pe*nth_loc
     do i=0,nr
     rmx=abs(den(i,j1,k)-dGr(i,j,k))
     dGr(i,j,k)=-(den(i,j1,k)+dGr(i,j,k))
    end do
    end do
   end do
  endif
 
  rmx=0.0
  rms=0.0
  rmx=maxval(abs(dGr(0:nr,0:nth_loc-1,0:nph-1)))
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    do i=0,nr
     if(dGr(i,j,k)==rmx)then
      mxind(1)=i;mxind(2)=j1
      mxind(3)=k
     endif
    end do
   end do
  end do
  den1=dGr*dGr
  rms=sum(den1(0:nr,0:nth_loc-1,0:nph-1))

 end subroutine div_err

 subroutine acc_res(err_mx,err_rms,lpe)
  integer,intent(in) :: lpe
  real,intent(inout) :: err_mx,err_rms
  real :: err1_max,err1_rms,dg_loc
  integer :: i,j,j1,k

  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+lpe*nth_loc
    do i=0,nr
     dg_loc=sqrt(dGph(i,j,k)**2+dGth(i,j,k)**2+dGr(i,j,k)**2)
     dGph(i,j,k)=dg_loc/sqrt(Gph(i,j1,k)**2+Gth(i,j1,k)**2+Gr(i,j1,k)**2)
    end do
   end do
  end do  
  err1_max=maxval(dGph(0:nr,0:nth_loc-1,0:nph-1))
  call allreduce_real_sdata(err1_max,err_mx,MAXV)
  den1=dGph*dGph
  err1_rms=sum(den1(0:nr,0:nth_loc-1,0:nph-1))
  call allreduce_real_sdata(err1_rms,err_rms,SUMV)
  err_rms=sqrt(err_rms/float(nr*nth*nph))

 end subroutine acc_res
!----------------------------
 subroutine field_energy
  integer :: i,j,k
  real :: emod2,emod

  do k=0,nph-1
   do j=0,nth-1
    do i=0,nr
     emod2=Gr(i,j,k)*Gr(i,j,k)+ & 
          Gth(i,j,k)*Gth(i,j,k)+ & 
          Gph(i,j,k)*Gph(i,j,k) 
     emod2=emod2/asq0
     emod=sqrt(emod2)
     aux(i,j,k)=asq0*(emod*sqrt(1.0+emod2)-log(emod+sqrt(1.0+emod2)))

    end do
   end do
  end do
!         now integrate the F(G/a0) function
  do j=0,nth-1
   do i=0,nr
    emod=sum(aux(i,j,0:nph-1))
    aux(i,j,0)=emod
   end do
  end do
   do i=0,nr
    emod=0.0
    do j=0,nth-1
     emod=emod+aux(i,j,0)*alp_inv(j,0,0)
    end do
    den1(i,0,0)=r(i,2)*emod
   end do
   epf=0.5*(den1(0,0,0)/r(0,3)+den1(nr,0,0)/r(nr,3))-den1(nr,0,0)
   do i=1,nr-1
    epf=epf+den1(i,0,0)/r(i,3)
   end do
   epf=0.5*epf/float(nph)
 end subroutine field_energy
!=================================
 subroutine mond_acc(loc_pe,loc_iter,itr)
 
  integer,intent(in) :: loc_pe,loc_iter
  integer,intent(out) :: itr
  real :: errg_max,errg_rms,g0_deep
  real :: rms,err_max,err1_rms,err1_max,loc_err
  integer :: i,j,j1,k,k1,k2,lmx,ml
  integer :: mind(3)
  integer :: itr_max,itr_min
  real,parameter :: small=1.e-08
  logical,parameter :: bcst=.true.

  rms=0.0
  itr=0
  tol=tol_ind*1.e-04
!--------------------------------
          ! enters the previous solution or an initial guess
          !as (aux,dGr,dGth,dGph)
!+++++++++++++++++++++++++++++++++++
  lmx=lmax
  ml=min(lmx,hnph-1)
  g0_deep=1.
  if(mond_ind==2)g0_deep=g_0
  

!++))+++++++++++++++++++++++++++++++++++++
 if(loc_iter==0)call mond_init(loc_pe)
                                        !At t=0 initializes with spherical
                                         !symmetric Mond solution
!--------------------------
   do k=0,nph-1
    do j=0,nth_loc-1
     j1=j+loc_pe*nth_loc
     do i=0,nr
      Gr(i,j1,k)=dGr(i,j,k)
      Gth(i,j1,k)=dGth(i,j,k)
      Gph(i,j1,k)=dGph(i,j,k)
     end do
    end do
   end do  
!-------------------------
                     ! iterative procedure 
  do
   itr=itr+1

   call Gmond(loc_pe,mond_ind)
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      dGr(i,j,k)=Gr(i,j1,k)*mu(i,j,k)
      dGth(i,j,k)=Gth(i,j1,k)*mu(i,j,k)
      dGph(i,j,k)=Gph(i,j1,k)*mu(i,j,k)
     end do
    end do
   end do  

  call divG(loc_pe) ! exit dGr=r*div(mu*G)

  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    do i=0,nr
     dGr(i,j,k)=-(g0_deep*den(i,j1,k)+&
                                 dGr(i,j,k))
    end do
   end do
  end do
!======================
  !call div_err(0,loc_pe,err_max,rms,mind)
!+++++++++++++++++++++++++++

 call psolv(dt_iter,loc_pe,lmx,ml)
  do k=0,nph-1
   do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
    do i=0,nr
     Gr(i,j1,k)=Gr(i,j1,k)+dGr(i,j,k)
     Gth(i,j1,k)=Gth(i,j1,k)+dGth(i,j,k)
     Gph(i,j1,k)=Gph(i,j1,k)+dGph(i,j,k)
    end do
   end do
  end do  
  call acc_res(errg_max,errg_rms,loc_pe)
  if(errg_rms <= tol.or.itr >=itermax)exit
!----------------------------
 end do
!++++++++++++++++++++++++++++++
  call div_err(1,loc_pe,err_max,rms,mind)
  if(itr==itermax)then
   if(pe0)write(99,*)'warning : max number of iterations used ! '
  endif
  call allreduce_real_sdata(err_max,err1_max,MAXV)
  call allreduce_real_sdata(rms,err1_rms,SUMV)
  err1_rms=sqrt(err1_rms/float(nr*nth*nph))
  call allreduce_int_sdata(itr,itr_max,MAXV)
  call allreduce_int_sdata(itr,itr_min,MINV)

  if(mod(loc_iter,mrate)==0)then
   if(loc_pe==0)then
    write(99,*)' Max/Min  local iteration numbers= ',itr_max,itr_min
    write(99,'(a11,e11.4)')' div_max   ',err1_max
    write(99,'(a11,e11.4)')' div_rms=  ',err1_rms
    write(99,'(a7,e11.4)')' dG_max  ',errg_max
    write(99,'(a7,e11.4)')' dG_rms  ',errg_rms
    write(99,'(a6,e11.4,a8,e11.4)')'time= ',tnow,' dtime= ',dtime
   endif
  endif
!+++++++++++++++++++
  call div_max(loc_pe)
                      !Computes div(G)_max to set stable time step 
!++++++++++++++++++++++++++++++++++++++++++++ 
                    ! sets a local storage of grid acceleration
                    ! for next iteration 
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      dGr(i,j,k)=Gr(i,j1,k)
      dGth(i,j,k)=Gth(i,j1,k)
      dGph(i,j,k)=Gph(i,j1,k)
     end do
    end do
   end do  
!++++++++++++++++++++++++++++++++++++++++
  if(prl)then
   aux=0.0
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      aux(i,j1,k)=dGr(i,j,k)
     end do
    end do
   end do  
   call sumreduce_grdata(bcst,aux,aux1,&
                         nr+1,nth,nph)
   Gr(0:nr,0:nth-1,0:nph-1)=aux1(0:nr,0:nth-1,0:nph-1)
   aux=0.0
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      aux(i,j1,k)=dGth(i,j,k)
     end do
    end do
   end do  
   call sumreduce_grdata(bcst,aux,aux1,&
                         nr+1,nth,nph)
   Gth(0:nr,0:nth-1,0:nph-1)=aux1(0:nr,0:nth-1,0:nph-1)
   aux=0.0
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pe*nth_loc
     do i=0,nr
      aux(i,j1,k)=dGph(i,j,k)
     end do
    end do
   end do  
   call sumreduce_grdata(bcst,aux,aux1,&
                         nr+1,nth,nph)
   Gph(0:nr,0:nth-1,0:nph-1)=aux1(0:nr,0:nth-1,0:nph-1)
  endif

  call field_energy
!====================================
! Sets G=[d/dr,d/dtheta,d/dphi]Pot=(1, r, rsin(theta)]*[Gr,Gth,Gph]^t
!--------------
   do k=0,nph-1
    do j=0,nth-1
     do i=0,nr
      Gth(i,j,k)=r(i,1)*Gth(i,j,k)
      Gph(i,j,k)=r(i,1)*sth(j)*Gph(i,j,k)
     end do
    end do
   end do

   call set_fbd(1)

 end subroutine mond_acc 
!----------------------------
 subroutine mond_pot(loc_pr,lm)
  integer,intent(in) :: loc_pr,lm
  integer :: ml,i,j,k,j1
  logical,parameter :: bst=.true.

  ml=min(lm,nph/2-1)
  !++++++++++++++++++++++++++++++++++++
  ! Uses the local acceleration arrays (dGr,dGth,dGph) previously stored
  !======================================

  call divG(loc_pr) ! exit dGr=r*div(G)
  den1(0:nr,0:nth_loc-1,0:nph-1)=dGr(0:nr,0:nth_loc-1,0:nph-1)
  call ft1_ph(den1,nr,nth_loc,nph,-1)
  call lgd_trsf(lm,ml,-1,loc_pr) 
  call newt_pot_inv(lm,ml,2)
   if(prl)then
    call sumreduce_grdata(bst,aux,aux1,&
                         nr+1,nth,nph)
    aux(0:nr,0,0)=aux1(0:nr,0,0)
   endif 
  call lgd_trsf(lm,ml,1,loc_pr)        
  call ft1_ph(den1,nr,nth_loc,nph,1)
!===================
         ! den1(0:nr,0:nth_loc-1,0:nph-1) is the potential in loc grid space 
!=====================
   do k=0,nph-1
    do j=0,nth_loc-1
    j1=j+loc_pr*nth_loc
     aux(0:nr,j1,k)=den1(0:nr,j,k)
    end do
   end do  
  if(prl)then
   call fill_grdata(aux,den1,dGr,nr,nth,nth_loc,nph)
  end if
!-----------------------------
       !aux(0:nr,0:nth-1,0:nph-1) is the potential in grid space 
       ! common to all Pe's 
!=========================
! Reset the original local acceleration (dGr,dGth,dGph) 
  do k=0,nph-1
   do j=0,nth_loc-1
   j1=j+loc_pr*nth_loc
    do i=0,nr
     dGr(i,j,k)=Gr(i,j1,k)
     dGth(i,j,k)=Gth(i,j1,k)/r(i,1)
     dGph(i,j,k)=Gph(i,j1,k)/(r(i,1)*sth(j1))
    end do
   end do
  end do  
 end subroutine mond_pot
!-------------------------
 subroutine set_fbd(is)
  integer,intent(in) :: is

  integer :: i,j,k,j1,k1
  real :: fz

!--------------------- boundary points around theta=0, theta=pi axis
 do k=0,nph-1
  k1=k+hnph
  if(k1 >(nph-1))k1=k1-nph
  Gr(0:nr,-1,k)=Gr(0:nr,0,k1)
  Gr(0:nr,nth,k)=Gr(0:nr,nth-1,k1)
  Gth(0:nr,-1,k)=-Gth(0:nr,0,k1)
  Gth(0:nr,nth,k)=-Gth(0:nr,nth-1,k1)
  Gph(0:nr,-1,k)=is*Gph(0:nr,0,k1)
  Gph(0:nr,nth,k)=is*Gph(0:nr,nth-1,k1)
 end do
!--------------------------------------
                        ! periodicity
  Gr(0:nr,-1:nth,nph)=Gr(0:nr,-1:nth,0)
  Gth(0:nr,-1:nth,nph)=Gth(0:nr,-1:nth,0)
  Gph(0:nr,-1:nth,nph)=Gph(0:nr,-1:nth,0)

  Gr(-1,-1:nth,0:nph)=Gr(0,-1:nth,0:nph)
  Gth(-1,-1:nth,0:nph)=Gth(0,-1:nth,0:nph)
  Gph(-1,-1:nth,0:nph)=Gph(0,-1:nth,0:nph)

 end subroutine set_fbd
!+++++++++++++++++++++++++++++++++++
 subroutine set_potbd

  integer :: i,j,k,j1,k1
  real :: fz
!--------------------- boundary points
                       !j=-1 > theta=-dth/2,j=nth => pi+dth/2
  do k=0,nph-1
   k1=k+hnph
   if(k1 > nph-1)k1=k1-nph
   Gr(0:nr,-1,k)=Gr(0:nr,0,k1)
   Gr(0:nr,nth,k)=Gr(0:nr,nth-1,k1)
  end do
  Gr(0:nr,-1:nth,nph)=Gr(0:nr,-1:nth,0)
  Gr(-1,-1:nth,0:nph)=Gr(0,-1:nth,0:nph)

 end subroutine set_potbd

end module work

! 28/07/2008 corrected typos: 
! now indices in dot_products in subroutins Rad_der and full_gradP are correct
! 1/6/2011  corrected index in full_lgd_trsf subroutine case(1)
