!
!   Copyright (C) 2007 by Pasquale Londrillo & Carlo Nipoti               
!   pasquale.londrillo@oabo.inaf.it   carlo.nipoti@unibo.it   
!                                                                         
!   This program is free software; you can redistribute it and/or modify  
!   it under the terms of the GNU General Public License 
!                                                                         
!   This program is distributed in the hope that it will be useful,       
!   but WITHOUT ANY WARRANTY; without even the implied warranty of        
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
!   GNU General Public License for more details.                          
!  
!   Indications of bugs identification and suggestions on possible 
!   improvement are welcome.  
!
!++++++++++++++++++++++++++++++++++++++++++++++++++++++
! This module contains the 'main' program routine 
! calling, in turn:
!=====================
! 1) the initializing routines:
! - " start_parallel" for MPI calls,
! - " start  " reading and setting the main constant parameters,
!              and particles dynamical variables,
!   " init"    computing field (grid-based) dynamical variables at 
!              the initial T=0 time;     
!================================
! 2) the "advance_in_time" routine, containing:
! -  "pos_adv()     "call, to move particle (x,y,z) cartesian positions

! -  "set_new_acc() "call, to compute new particle accelerations. 
!                          It contains the sub-steps:
!    a) "set_den()" call,  to assign particle mass-weights to a spherical grid; 
!    b) "set_gr_den"call,  to set a filtered source term for Poisson equation;
!    c) "mond_acc()" or "newt_acc" call, 
!                        to solve for (potential-acceleration) grid-fields; 
!    d) "part_pot()" and "part_acc()" calls, 
!                        to interpolate the potential-acceleration fields 
!                        at each particle position.

! -  "vel_adv()   " call, to move particle velocities.
!======================
! 3) The output routine "data_out", containing:
!-   "part_output" call, to write particles data, in binary form
!-   "pot_data" call, to write particle potential, in binary form
!-   "grid_bdata" call, to write grid-based field data, in binary form
!    "endata" call, to write integrated dynamical variables, 
!     previously stored in "diags" calls, in ASCII form
!====================================================
!====================================================
!====================================================
 program mond3D

  use work 
  use particles 
  use setio

  implicit none

  integer :: iter,id_new,new,model,lp_ord
  real :: lpf(3)
  real(kind=8) :: t1,t2,cptot,cpgr
!===========================

   call start_parallel

   call start 

   call init(model)
!===========================
   iter=0
   cptot=0.0
   cpgr=0.0
   if(model >0 )then
     call end_run
     stop
   endif
!-----------------------
   if(new==0)call init_data_out
!-----------------------
   cptot=0.0
   call second(t1)

  if(tmax> 0.0)then
   do 
    call advance_in_time(iter,tnow) 

    iter=iter+1
    call data_out
    call second(t2)
    cptot=cptot+t2-t1
    if(mod(iter,mrate)==0)then
     if(pe0)then
      write(99,'(a6,i5,a5,e11.4)')'step= ',iter,' dt= ',dtime
      write(99,'(a6,e11.4)')'time= ',tnow
      write(99,'(a12,2e11.4)')' cpu_times =',cptot,cpgr
     endif
    endif
    call second(t1)
    if(tnow >=tmax)exit
   end do
  endif 
  if(pe0)then
    write(99,'(a20,i5,a7,e11.4)')' end of run:  iter= ',iter,' time= ',tnow
    write(99,'(a9,e11.4,a11,e11.4)')'tot_cpu= ',cptot,' grid_cpu= ',cpgr
    write(99,*)'END OF RUN'
  end if
  call end_run
  
  contains
  subroutine start

   open(10,file='input.data')

   read(10,*)
   read(10,*)nr,nth,nph,lmax
   read(10,*)
   read(10,*)model,mond_ind,spl_ord,rmap
   read(10,*)
   read(10,*)g_0,dt_iter,rad_sc
   read(10,*)
   read(10,*)itermax,tol_ind
   read(10,*)
   read(10,*)new,id_new,nout,iene,mrate
   read(10,*)
   read(10,*)tmax,dt_min,cfl,lp_ord

  close(10)

  asq0=g_0*g_0 
  iout=id_new
  rder=1         !Selects, by default, a FD scheme for Laplacian discretization
  nph_loc=nph/npe
  nth_loc=nth/npe
  if(prl)allocate(sbuff(nr+1,nph,2))
  call v_alloc(nr,nth,nth_loc,nph)
  call mainv_alloc(nr,nth,nph,mem_size)
!  call ftw_init(nph)
  call set_rthph_grid
  call alv_alloc(nr,nth,nph)
  call set_rad_mat       
  call set_leg
  call set_poisson_mat(lmax)
!-------------- 4th leap-frog coefficients
  lpf=1.0
  if(lp_ord==4)then
   lpf(1)=1.0/(2.0-2.0**(1.0/3.0))
   lpf(2)=1.0-2.0*lpf(1)
   lpf(3)=lpf(1)
  endif 
  if(model>0)tmax=0.0 
  part_size=0
!--------------------
 if(model==0)then
  call part_init(iout,part_size)
!----------- Parameters as Common data for nbd_lib routines
  rm=rmap
  sc=rad_sc
  ord=spl_ord
  rx=dx1
  ry=dth1
  rz=dph1
  pig=acos(-1.0)
  nb=nbd(mype)
 endif
!--------------------------------------
  if(pe0)call run_info

!  call check_ft

 end subroutine start
!================================
 subroutine run_info
  write(99,*)'start parallel, nproc=',npe
  write(99,*)'Input grid parameters'
  write(99,*)'  nr , nth, nph, lmax'
  write(99,'(4i5)')nr,nth,nph,lmax
  write(99,*)'Radial grid Min-Max'
  write(99,'(2e11.4)')r(0,1),r(nr,1)
  if(model==0)then
   write(99,*)'Tot particle number'
   write(99,'(a1,i8)')' ',nbd_tot
   mem_size=mem_size+part_size
  endif
  if(mond_ind==0)write(99,*)'RUNs Newtonian Gravity'
  if(mond_ind==1)write(99,*)'RUNs Mond  Gravity'
  if(mond_ind==2)write(99,*)'RUNs Deep-Mond Gravity'
  write(99,*)'+++++++++++++++++++++++++++++'
  mem_size=mem_size*kind(1.0)
  write(99,'(a17,i10)')'Allocated bytes= ',mem_size
  write(99,'(a10,i3)')'data_kind ',kind(1.0)
  write(99,*)'+++++++++++++++++++++++++++++'
 end subroutine run_info
!===========================================
 subroutine check_gr_mass
  real:: gmass
  integer :: i,j,k
                        !checks for total mass after filtering
  gmass=0.0
   do k=0,nph-1
    do j=0,nth-1
     do i=0,nr
      gmass=gmass+aux(i,j,k)*drv(i,j)
     end do
    end do
   end do
   write(99,'(a16,e11.4)')'Check grid mass=',gmass
  
 end subroutine check_gr_mass
!=================================
 subroutine set_gr_den

 integer :: i,j,lm,ipe
 logical :: sr
 real :: gmass,msum

 den(0:nr,0:nth-1,0:nph-1)=pmass*den(0:nr,0:nth-1,0:nph-1)
 aux(0:nr,0:nth-1,0:nph-1)=den(0:nr,0:nth-1,0:nph-1)
!=====================================
 if(prl)then
  sr=.true.
  aux1(0:nr,0:nth-1,0:nph-1)=den(0:nr,0:nth-1,0:nph-1)
  call sumreduce_grdata(sr,aux1(0:nr,0:nth-1,0:nph-1),&
                              aux(0:nr,0:nth-1,0:nph-1),nr+1,nth,nph)
 endif
 
  do i=0,nr
   den1(i,0,0)=sum(aux(i,0:nth-1,0:nph-1))
  end do
  msum=sum(den1(0:nr,0,0))
  gmass=0.0
  i=0 
  do 
   gmass=gmass+den1(i,0,0)
   if(gmass > 0.5*msum)then
    gmass=gmass-den1(i,0,0)
    j=i-1
    rh=r(j,1)+(r(j+1,1)-r(j,1))*(0.5*msum-gmass)/den1(j,0,0)
    exit
   endif
   i=i+1
  end do
!----------------------------   
  do j=0,nth-1
   do i=0,nr
    aux(i,j,0:nph-1)= & 
              aux(i,j,0:nph-1)/drv(i,j)
   end do
  end do
  dmax=maxval(aux(0:nr,0:nth-1,0:nph-1))

  if(lmax < nth-1)call filt(lmax)
!=======================
 do i=0,nr
  den(i,0:nth-1,0:nph-1)=gravc*r(i,1)*aux(i,0:nth-1,0:nph-1)
 end do
!---------------------------------------------
!Exit  den()=4Pi*G*r*rho ithe source term in Poisson's equation
!================================
 end subroutine set_gr_den

 subroutine init(modl)
  integer,intent(in) :: modl

  integer :: i,j,k,lm,loc_iter,miter,loc_ite(1)
   
!---------------------
  loc_iter=0
  if(modl >0)then
   call gmodel(modl)
   if(pe0)then
    call grid_bdata(1,4)
    write(99,*)' static model completed' 
   endif
   return
  endif
!-----------------------------
                          ! assigns particle weights to a spherical grid
  call set_den(nr,nth,nph)
                          !sets a filtered source term for Poisson solver:
                          ! den=4*Pi*G*r*rho 
  call set_gr_den
  if(pe0)call check_gr_mass

!------------------------------------
  if(mond_ind >0)then
   call mond_acc(mype,loc_iter,miter)
   call mond_pot(mype,lmax)
  else
   call newt_acc(mype,loc_iter)
  endif
  dtime=cfl/sqrt(divg_max)
  dth=0.5*dtime
!--------------------------------------
   dt_out=tmax/float(nout)
   dt_ene=tmax/float(iene)
   te=tnow
   tout=tnow
   nte=0
   ite=0
   if(new==1)then
    tmax=tnow+tmax
    if(pe0)call read_en_data(iout,loc_ite(1))
    call bcast_int_data(1,loc_ite) 
    ite=loc_ite(1)
    te=te+dt_ene
    tout=tout+dt_out
   endif
 end subroutine init
!----------------------------------
 subroutine gmodel(gm)
  integer,intent(in) :: gm
  integer :: i,j,k,loc_iter,miter
  real :: rp,xi,xx,yy,zz,a,b,b2,fact,pot0

  select case(gm)

  case(1)
   do i=0,nr
    den(i,0:nth-1,0:nph-1)=2.0/(1.0+r(i,1))**3    !=4Pi*G*r*rho the source term
    aux(i,0:nth-1,0:nph-1)=1.0/(1.0+r(i,1))    !=4Pi*G*r*rho the source term
    Gr(i,0:nth-1,0:nph-1)=-1.0/(1.0+r(i,1))**2    !=4Pi*G*r*rho the source term
   end do
   Gph=0.0
   Gth=0.0
   case(2)
   !call check_thder

   a=1.0;b=0.3
   b2=b*b

   do j=0,nth-1
    do i=0,nr
     rp=r(i,1)*sth(j)
     zz=r(i,1)*cth(j)
     xx=sqrt(zz*zz+b2)
     xi=a+xx
     fact=rp*rp+xi*xi
     pot0=1.0/(fact*sqrt(fact))
     yy=a/xx
     aux(i,j,0:nph-1)=1.0/sqrt(fact)
     Gr(i,j,0:nph-1)=-r(i,1)*pot0*(&
                    sth(j)*sth(j)+(1.0+yy)*cth(j)*cth(j))
     Gth(i,j,0:nph-1)=r(i,1)*r(i,1)*pot0*cth(j)*sth(j)*yy
     den(i,j,0:nph-1)=b2*(a*rp*rp+(a+3.*xx)*xi*xi)
     den(i,j,0:nph-1)=r(i,1)*den(i,j,0:nph-1)/(xx*xx*xx*fact**(2.5))
    end do
   end do
   Gph=0.0
   case(3)

   do j=0,nth-1
    do i=0,nr
     den(i,j,0:nph-1)=-1.0/(1.+r(i,1))
     Gth(i,j,0:nph-1)=2./(1.+r(i,1))**3
    end do
   end do
   call Check_Lapl(mype)
   write(20,*)'Density potential pair'
   do i=0,nr
    write(20,'(2e11.4)')den1(i,0,0),Gth(i,0,0)-den1(i,0,0)
   end do
   return
  end select
!------------------
  call grid_bdata(0,4)
  loc_iter=0
   
   if(mond_ind >0)then
    call mond_acc(mype,loc_iter,miter)
    if(pe0)write(99,*)' mond iterations = ' ,miter
   else
    call newt_acc(mype,loc_iter)
   endif
 end subroutine gmodel

 subroutine set_new_acc(modl)

  integer,intent(in) :: modl
  real(kind=8)cp1,cp2
   logical :: sr
   integer :: loc_iter,miter

  loc_iter=0 
  if(modl==0)then
   loc_iter=iter

   call set_den(nr,nth,nph) 

   call set_gr_den         ! stores dmax 

  endif
   call second(cp1)

   if(mond_ind >0)call mond_acc(mype,loc_iter,miter)
   if(mond_ind ==0)call newt_acc(mype,loc_iter)

   call second(cp2)
   cpgr=cpgr+cp2-cp1
   !write(99,*)'Psolv steps',loc_iter
   !write(99,*)'Psolv time',cp2-cp1

   call part_acc(nr,nth,nph)

 end subroutine set_new_acc 
!----------------------------
 subroutine advance_in_time(it,t_loc) 

  integer,intent(in) :: it
  real,intent(inout) :: t_loc
  real :: dt_loc,dth_loc
  integer :: ik

  dtime=cfl/sqrt(divg_max)
  If(dtime < dt_min)then
   if(pe0)write(99,*)'Warning, too small a time step'
   dtime=dt_min
  endif
!==========================================
  do ik=1,lp_ord-1
   dt_loc=lpf(ik)*dtime
   hdt=0.5*dt_loc
    
   call pos_adv(hdt)

   call set_new_acc(0)

   call vel_adv(dt_loc)

   call pos_adv(hdt)
  end do
  t_loc=t_loc+dtime
!==========================================
  end subroutine advance_in_time
!-------------------------------------
 subroutine init_data_out 

   if(pe0)call grid_bdata(iout,nph)
   call part_acc(nr,nth,nph)

   Gr(0:nr,0:nth-1,0:nph-1)=aux(0:nr,0:nth-1,0:nph-1) 
   call set_potbd 
   call part_pot(Gr,nr,nth,nph)
   call diags
   call pot_data
   tout=tout+dt_out
   if(pe0)then
    call endata(ite,nte,iout)
   endif
 end subroutine init_data_out
!----------------------
 subroutine data_out 

  if(tnow >=te.or.tnow>=tout)then
   if(mond_ind >0)call mond_pot(mype,lmax)
   if(tnow>=tout)then
    call part_output
    if(pe0)call grid_bdata(iout,nph)
   endif
   Gr(0:nr,0:nth-1,0:nph-1)=aux(0:nr,0:nth-1,0:nph-1) 
   call set_potbd 
   call part_pot(Gr,nr,nth,nph)
   if(tnow>=te)call diags
    if(tnow>=tout)then
     call pot_data
     tout=tout+dt_out
     if(pe0)then
      call endata(ite,nte,iout)
     endif
    endif
   endif
  end subroutine data_out
  subroutine end_run

   call all_dealloc
   call end_parallel

  end subroutine end_run

  subroutine second(stime)

   real(kind=8):: stime
   integer :: rate,irate

   call system_clock(rate,irate)
   stime=float(rate)/float(irate)
 end subroutine second
 

 end program mond3D
!----------------------------
!last change 23 jan 2008
!------------------------------
